# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# BuildRPMDriver derived classes expose the following methods:
#
# build_package(). This method will perform the actual package build using
#                  the driver-specific approach.

from dlrn.drivers.buildrpm import BuildRPMDriver
import dlrn.shell

import io
import logging
import multiprocessing
import os
import re
import sh

from time import localtime
from time import strftime
from time import time

logger = logging.getLogger("dlrn-build-koji")


class KojiBuildDriver(BuildRPMDriver):
    DRIVER_CONFIG = {
        'kojibuild_driver': {
            'koji_krb_principal': {'name': 'krb_principal'},
            'koji_krb_keytab': {'name': 'krb_keytab'},
            'koji_scratch_build': {'name': 'scratch_build', 'type': 'boolean',
                                   'default': True},
            'koji_build_target': {'name': 'build_target'},
            'koji_arch': {'name': 'arch', 'default': 'x86_64'},
            'koji_use_rhpkg': {'name': 'use_rhpkg', 'type': 'boolean'},
            'koji_exe': {'default': 'koji'},
            'fetch_mock_config': {'type': 'boolean'},
            'mock_base_packages': {'default': ''},
            'mock_package_manager': {'default': ''},
            'koji_add_tags': {'name': 'additional_koji_tags', 'type': 'list',
                              'default': []},
        }
    }

    def __init__(self, *args, **kwargs):
        super(KojiBuildDriver, self).__init__(*args, **kwargs)
        self.exe_name = self.config_options.koji_exe
        # Check for empty additional_koji_tags value
        if self.config_options.koji_add_tags == ['']:
            self.config_options.koji_add_tags = []

    # We are using this method to "tee" koji output to a log file and stdout
    def _process_koji_output(self, line):
        if dlrn.shell.verbose_build:
            logger.info(line[:-1])
        self.koji_fp.write(line)

    def write_mock_config(self, filename):
        """Retrieve mock config from Koji instance

        :param filename: output filename to write mock config
        """
        target = self.config_options.koji_build_target
        arch = self.config_options.koji_arch
        try:
            worker_id = multiprocessing.current_process()._identity[0]
        except IndexError:
            # Not in multiprocessing mode
            worker_id = 1
        run_cmd = [self.exe_name]
        run_cmd.extend(['mock-config',
                        '--arch', arch, '--target', target, '-o', filename])
        # FIXME(hguemar): add proper exception management
        sh.env(run_cmd,
               _env={'PATH': '/usr/bin/'})
        lines = []
        with open(filename, 'r') as fp:
            for line in fp.readlines():
                if (line.startswith("config_opts['chroot_setup_cmd']") and
                        self.config_options.mock_base_packages != ''):
                    lines.append("config_opts['chroot_setup_cmd'] = "
                                 "'install %s'\n" %
                                 self.config_options.mock_base_packages)
                elif line.startswith("config_opts['root']"):
                    # Append worker id to mock buildroot name
                    line = line[:-2] + "-" + str(worker_id) + "'\n"
                    lines.append(line)
                else:
                    lines.append(line)
            if self.config_options.mock_package_manager:
                lines.append("config_opts['package_manager'] = '%s'\n" %
                             self.config_options.mock_package_manager)
        with open(filename, 'w') as fp:
            fp.write(''.join(lines))

    def _build_with_rhpkg(self, package_name, output_dir, src_rpm, scratch,
                          commit):
        """Use rhpkg as build backend

        :param package_name: package name to build
        :param output_dir: output directory
        :param src_rpm: source RPM to build
        :param scratch: define if build is scratch or not
        """
        distgit_dir = os.path.join(
            self.config_options.datadir,
            package_name + "_distro")

        build_exception = None

        # if we are using rhpkg, we need to create a kerberos ticket
        krb_principal = self.config_options.koji_krb_principal
        keytab_file = self.config_options.koji_krb_keytab
        with io.open("%s/kerberos.log" % output_dir, 'a',
                     encoding='utf-8', errors='replace') as self.koji_fp:
            sh.kinit('-k', '-t', keytab_file, krb_principal)

        rhpkg = sh.rhpkg.bake(_cwd=distgit_dir, _tty_out=False,
                              _timeout=3600,
                              _err=self._process_koji_output,
                              _out=self._process_koji_output,
                              _env={'PATH': '/usr/bin/'})

        if (self.config_options.pkginfo_driver ==
            'dlrn.drivers.downstream.DownstreamInfoDriver' and
                self.config_options.use_upstream_spec):
            # This is a special situation. We are copying the upstream
            # spec over, but then building the srpm and importing. In this
            # situation, rhpkg import will complain because there are
            # uncommited changes to the repo... and we will commit them
            # the srpm. So let's reset the git repo right before that.
            git = sh.git.bake(_cwd=distgit_dir, _tty_out=False,
                              _timeout=3600,
                              _err=self._process_koji_output,
                              _out=self._process_koji_output,
                              _env={'PATH': '/usr/bin/'})
            git.checkout('--', '*')

        with io.open("%s/rhpkgimport.log" % output_dir, 'a',
                     encoding='utf-8', errors='replace') as self.koji_fp:
            rhpkg('import', '--skip-diff', src_rpm)
            # Get build NVR
            m = re.search(r'([0-9a-zA-Z._+-]+)\.src\.rpm', src_rpm)
            if m and m.group(1):
                package_nvr = m.group(1)
            else:
                package_nvr = 'XXX-XXX'

            pkg_date = strftime("%Y-%m-%d-%H%M%S", localtime(time()))
            rhpkg('commit', '-p', '-m',
                  'DLRN build at %s\n\nSource SHA: %s\nDist SHA: %s\n'
                  'NVR: %s\n' %
                  (pkg_date, commit.commit_hash, commit.distro_hash,
                   package_nvr))

        # After running rhpkg commit, we have a different commit hash, so
        # let's update it
        git = sh.git.bake(_cwd=distgit_dir, _tty_out=False, _timeout=3600)
        repoinfo = str(git.log("--pretty=format:%H %ct", "-1", ".")).\
            strip().split(" ")

        logger.info("Updated git: %s" % repoinfo)
        # When using rhpkg with a pkginfo driver other than downstreamdriver,
        # we want to overwrite the distro_hash instead of extended_hash.
        # Otherwise, The distgit update will trigger yet another build on
        # the next run, causing an endless loop
        if (self.config_options.pkginfo_driver !=
                'dlrn.drivers.downstream.DownstreamInfoDriver'):
            commit.distro_hash = repoinfo[0]
            commit.dt_distro = repoinfo[1]
        else:
            commit.extended_hash = repoinfo[0]
            commit.dt_extended = repoinfo[1]

        # Since we are changing the extended_hash, we need to rename the
        # output directory to match the updated value
        datadir = os.path.realpath(self.config_options.datadir)
        new_output_dir = os.path.join(datadir, "repos",
                                      commit.getshardedcommitdir())
        logger.info("Renaming %s to %s" % (output_dir, new_output_dir))
        os.rename(output_dir, new_output_dir)
        output_dir = new_output_dir

        with io.open("%s/rhpkgbuild.log" % output_dir, 'a',
                     encoding='utf-8', errors='replace') as self.koji_fp:
            try:
                # NOTE(jpena): we are using --skip-nvr-check to avoid
                # issues when building RHEL8 packages on RHEL7, due to
                # new RPM features.
                rhpkg('build', '--skip-nvr-check', scratch=scratch)
            except Exception as e:
                build_exception = e

        return build_exception, "%s/rhpkgbuild.log" % output_dir

    def _build_with_exe(self, package_name, output_dir, src_rpm, scratch,
                        commit):
        """Build using koji/brew executables (cbs being an aliases)

        :param package_name: package name to build
        :param output_dir: output directory
        :param src_rpm: source RPM to build
        :param scratch: define if build is scratch or not
        """

        krb_principal = self.config_options.koji_krb_principal
        keytab_file = self.config_options.koji_krb_keytab
        scratch = self.config_options.koji_scratch_build
        target = self.config_options.koji_build_target

        # Build package using koji/brew
        run_cmd = [self.exe_name]
        if krb_principal:
            run_cmd.extend(['--principal', krb_principal,
                            '--keytab', keytab_file])
        run_cmd.extend(['build', '--wait',
                        target, src_rpm])

        build_exception = None
        with io.open("%s/kojibuild.log" % output_dir, 'a',
                     encoding='utf-8', errors='replace') as self.koji_fp:
            try:
                sh.env(run_cmd, _err=self._process_koji_output,
                       _out=self._process_koji_output,
                       _cwd=output_dir, scratch=scratch,
                       _env={'PATH': '/usr/bin/'})
            except Exception as e:
                build_exception = e

        return build_exception, "%s/kojibuild.log" % output_dir

    def build_package(self, **kwargs):
        """Valid parameters:

        :param output_directory: directory where the SRPM is located,
                                 and the built packages will be.
        :param package_name: name of a package to build
        """
        output_dir = kwargs.get('output_directory')
        package_name = kwargs.get('package_name')
        commit = kwargs.get('commit')
        scratch = self.config_options.koji_scratch_build
        build_exception = None

        # Find src.rpm
        for rpm in os.listdir(output_dir):
            if rpm.endswith(".src.rpm"):
                src_rpm = os.path.realpath('%s/%s' % (output_dir, rpm))
        try:
            if self.config_options.koji_use_rhpkg:
                build_method = self._build_with_rhpkg
            else:
                build_method = self._build_with_exe
            build_exception, logfile = build_method(
                package_name, output_dir, src_rpm, scratch, commit)

            if self.config_options.koji_use_rhpkg:
                # In this case, we need to re-calculate the output directory
                datadir = os.path.realpath(self.config_options.datadir)
                output_dir = os.path.join(datadir, "repos",
                                          commit.getshardedcommitdir())

            # Find task id to download logs
            with open(logfile, 'r') as fp:
                log_content = fp.readlines()
            task_id = None
            for line in log_content:
                m = re.search(r'^Created task: (\d+)$', line)
                if m:
                    logger.info("Created task id %s" % m.group(1))
                    task_id = m.group(1)
                    break

            if not task_id:
                raise Exception('Failed to find task id for the koji build')

            # Also find package name if we need to add tags
            if len(self.config_options.koji_add_tags) > 0:
                # Get build name
                m = re.search(r'([0-9a-zA-Z._+-]+)\.src\.rpm', src_rpm)
                package_nvr = None
                if m:
                    logger.info("Adding tags for %s" % m.group(1))
                    package_nvr = m.group(1)
                if not package_nvr:
                    raise Exception('Failed to find package nvr when tagging')

                for tag in self.config_options.koji_add_tags:
                    run_cmd = []
                    run_cmd.extend(
                        [self.exe_name, 'tag-build', tag, package_nvr])

                    with io.open("%s/additional_tags.log" % output_dir, 'a',
                                 encoding='utf-8',
                                 errors='replace') as self.koji_fp:
                        try:
                            sh.env(run_cmd, _err=self._process_koji_output,
                                   _out=self._process_koji_output,
                                   _cwd=output_dir, _env={'PATH': '/usr/bin/'})
                        except Exception as e:
                            raise e

            # Download build artifacts and logs
            run_cmd = []
            run_cmd.extend(
                [self.exe_name,
                 'download-task', '--logs',
                 task_id])

            with io.open("%s/build_download.log" % output_dir, 'a',
                         encoding='utf-8', errors='replace') as self.koji_fp:
                try:
                    sh.env(run_cmd, _err=self._process_koji_output,
                           _out=self._process_koji_output,
                           _cwd=output_dir, _env={'PATH': '/usr/bin/'})
                except Exception as e:
                    raise e

            # All went fine, create the $OUTPUT_DIRECTORY/installed file
            open('%s/installed' % output_dir, 'a').close()
        finally:
            # Finally run restorecon
            try:
                sh.restorecon('-Rv', output_dir)
            except Exception as e:
                logger.info('restorecon did not run correctly, %s' % e)

            # We only want to raise the build exception at the very end, after
            # downloading all relevant artifacts
            if build_exception:
                raise build_exception



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
import os
import sh
import shutil
import tempfile

from six.moves import configparser

from dlrn.config import ConfigOptions
from dlrn.drivers.gitrepo import GitRepoDriver
from dlrn.tests import base


def _mocked_environ(*args, **kwargs):
    return 'myuser'


def _mocked_exists(path):
    return True


class TestDriverGit(base.TestCase):
    def setUp(self):
        super(TestDriverGit, self).setUp()
        config = configparser.RawConfigParser()
        config.read("projects.ini")
        config.set("DEFAULT", "pkginfo_driver",
                   "dlrn.drivers.gitrepo.GitRepoDriver")
        self.config = ConfigOptions(config)
        self.config.datadir = tempfile.mkdtemp()
        self.config.gitrepo_dirs = ['/openstack']

    def tearDown(self):
        super(TestDriverGit, self).tearDown()
        shutil.rmtree(self.config.datadir)

    @mock.patch.object(sh.Command, '__call__', autospec=True)
    @mock.patch('dlrn.drivers.gitrepo.refreshrepo')
    def test_getinfo(self, refresh_mock, sh_mock):
        refresh_mock.return_value = [None, None, None]
        driver = GitRepoDriver(cfg_options=self.config)
        package = {'upstream': 'test', 'name': 'test'}
        info = driver.getinfo(package=package, project="test", dev_mode=True)
        self.assertEqual(info, [])

    @mock.patch.object(sh.Command, '__call__', autospec=True)
    @mock.patch('os.listdir')
    def test_getpackages(self, listdir_mock, sh_mock):
        listdir_mock.return_value = []
        driver = GitRepoDriver(cfg_options=self.config)
        packages = driver.getpackages(dev_mode=True)
        self.assertEqual(packages, [])

    @mock.patch('os.path.exists', side_effect=_mocked_exists)
    @mock.patch('os.environ.get', side_effect=['myuser'])
    @mock.patch('sh.renderspec', create=True)
    @mock.patch('sh.env', create=True)
    @mock.patch('os.listdir')
    def test_custom_preprocess(self, ld_mock, env_mock, rs_mock, get_mock,
                               pth_mock):
        self.config.custom_preprocess = ['/bin/true']
        driver = GitRepoDriver(cfg_options=self.config)
        driver.preprocess(package_name='foo')

        directory = '%s/package_info/openstack/foo' % self.config.datadir

        expected = [mock.call(
            ['DLRN_PACKAGE_NAME=foo',
             'DLRN_DISTGIT=%s' % directory,
             'DLRN_SOURCEDIR=%s/foo' % self.config.datadir,
             'DLRN_USER=myuser',
             '/bin/true'],
            _cwd=directory,
            _env={'LANG': 'C'})]

        self.assertEqual(env_mock.call_args_list, expected)
        self.assertEqual(env_mock.call_count, 1)

    @mock.patch('os.path.exists', side_effect=_mocked_exists)
    @mock.patch('os.environ.get', side_effect=_mocked_environ)
    @mock.patch('sh.renderspec', create=True)
    @mock.patch('sh.env', create=True)
    @mock.patch('os.listdir')
    def test_custom_preprocess_multiple_commands(self, ld_mock, env_mock,
                                                 rs_mock, get_mock, pth_mock):
        self.config.custom_preprocess = ['/bin/true', '/bin/false']
        driver = GitRepoDriver(cfg_options=self.config)
        driver.preprocess(package_name='foo')

        directory = '%s/package_info/openstack/foo' % self.config.datadir

        expected = [mock.call(
            ['DLRN_PACKAGE_NAME=foo',
             'DLRN_DISTGIT=%s' % directory,
             'DLRN_SOURCEDIR=%s/foo' % self.config.datadir,
             'DLRN_USER=myuser',
             '/bin/true'],
            _cwd=directory,
            _env={'LANG': 'C'}),
            mock.call(
            ['DLRN_PACKAGE_NAME=foo',
             'DLRN_DISTGIT=%s' % directory,
             'DLRN_SOURCEDIR=%s/foo' % self.config.datadir,
             'DLRN_USER=myuser',
             '/bin/false'],
            _cwd=directory,
            _env={'LANG': 'C'})
            ]

        self.assertEqual(env_mock.call_args_list, expected)
        self.assertEqual(env_mock.call_count, 2)

    @mock.patch('sh.renderspec', create=True)
    @mock.patch('os.listdir')
    def test_custom_preprocess_fail(self, ld_mock, rs_mock):
        self.config.custom_preprocess = ['/bin/nonexistingcommand']
        driver = GitRepoDriver(cfg_options=self.config)
        os.makedirs(os.path.join(self.config.datadir,
                                 'package_info/openstack/foo'))

        self.assertRaises(RuntimeError, driver.preprocess, package_name='foo')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os

from flask import Flask

app = Flask(__name__)
app.config.from_object('dlrn.api.config')
try:
    app.config.from_pyfile(os.environ['CONFIG_FILE'], silent=True)
except KeyError:
    pass

from dlrn.api import dlrn_api  # noqa
from dlrn.config import setup_logging  # noqa


setup_logging(os.environ.get("DLRN_DEBUG"), os.environ.get("DLRN_LOG_FILE"))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import os
import time

import sqlalchemy
import sqlalchemy.ext.declarative

from sqlalchemy import asc
from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import desc
from sqlalchemy import event
from sqlalchemy import exc
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import or_
from sqlalchemy import String
from sqlalchemy import Text

from sqlalchemy.engine.base import Engine
from sqlalchemy.orm import scoped_session
from sqlalchemy.orm import synonym
from sqlalchemy.pool import NullPool

Base = sqlalchemy.ext.declarative.declarative_base()


class Commit(Base):
    __tablename__ = "commits"

    id = Column(Integer, primary_key=True)
    # Type has a default value for safety, this may be dropped in the future
    type = Column(String(18), default="rpm")
    dt_commit = Column(Integer)
    dt_distro = Column(Integer)
    dt_extended = Column(Integer)
    dt_build = Column(Integer)
    project_name = Column(String(256))
    repo_dir = Column(String(1024))
    distgit_dir = Column(String(1024))
    commit_hash = Column(String(64))
    distro_hash = Column(String(64))
    extended_hash = Column(String(64))
    commit_branch = Column(String(256))
    status = Column(String(64))
    component = Column(String(64))
    artifacts = Column(Text)
    notes = Column(Text)
    flags = Column(Integer, default=0)

    # For backwards compatibility when importing commits
    rpms = synonym("artifacts")

    def __gt__(self, b):
        return self.dt_commit > b.dt_commit

    def __lt__(self, b):
        return self.dt_commit < b.dt_commit

    def __eq__(self, b):
        return self.dt_commit == b.dt_commit

    def getshardedcommitdir(self):
        distro_hash_suffix = ""
        if self.distro_hash:
            distro_hash_suffix = "_%s" % self.distro_hash[:8]
        if self.extended_hash:
            # The extended hash is assumed to be a git-like hash, however
            # there is nothing preventing pkginfo drivers from using a
            # different format (such as YYYYMMDD).
            extended_hash_suffix = "_%s" % self.extended_hash[:8]
        else:
            extended_hash_suffix = ''
        if self.component is not None and self.component != 'None':
            component_prefix = 'component/%s/' % self.component
        else:
            component_prefix = ''

        hash_dir = self.commit_hash + distro_hash_suffix + extended_hash_suffix
        return os.path.join(component_prefix, self.commit_hash[:2],
                            self.commit_hash[2:4], hash_dir)


class Project(Base):
    __tablename__ = "projects"

    id = Column(Integer, primary_key=True)
    project_name = Column(String(256))
    last_email = Column(Integer)

    # Returns True if the last email sent for this project
    # was less then 24 hours ago
    def suppress_email(self):
        ct = time.time()
        if ct - self.last_email > 86400:
            return False
        return True

    def sent_email(self):
        self.last_email = int(time.time())


class CIVote(Base):
    __tablename__ = "civotes"

    id = Column(Integer, primary_key=True)
    commit_id = Column(Integer, ForeignKey('commits.id'), nullable=False)
    ci_name = Column(String(256))
    ci_url = Column(String(1024))
    ci_vote = Column(Boolean)
    ci_in_progress = Column(Boolean)
    timestamp = Column(Integer)
    notes = Column(Text)
    user = Column(String(255),
                  ForeignKey('users.username', name='civ_user_fk'))
    component = Column(String(64))


class CIVote_Aggregate(Base):
    __tablename__ = "civotes_agg"

    id = Column(Integer, primary_key=True)
    ref_hash = Column(String(64), nullable=False)
    ci_name = Column(String(256))
    ci_url = Column(String(1024))
    ci_vote = Column(Boolean)
    ci_in_progress = Column(Boolean)
    timestamp = Column(Integer)
    notes = Column(Text)
    user = Column(String(255),
                  ForeignKey('users.username'))


class Promotion(Base):
    __tablename__ = "promotions"

    id = Column(Integer, primary_key=True)
    commit_id = Column(Integer, ForeignKey('commits.id'), nullable=False)
    promotion_name = Column(String(256), nullable=False)
    timestamp = Column(Integer, nullable=False)
    user = Column(String(255),
                  ForeignKey('users.username', name='prom_user_fk'))
    component = Column(String(64))
    aggregate_hash = Column(String(64))


class User(Base):
    __tablename__ = "users"

    username = Column(String(255), primary_key=True)
    password = Column(String(256), nullable=False)


# Return a db session
def getSession(url='sqlite://'):
    engine = sqlalchemy.create_engine(url, poolclass=NullPool)
    Base.metadata.create_all(engine)
    Session = scoped_session(sqlalchemy.orm.sessionmaker(bind=engine))
    return Session


# Close a db session
def closeSession(session):
    session.remove()


# Get the most recently processed commit for project_name, we ignore commits
# with a status of "RETRY" as we want to retry these.
def getLastProcessedCommit(
        session, project_name, not_status="RETRY", type='rpm'):
    commit = session.query(Commit).filter(Commit.project_name == project_name,
                                          Commit.type == type,
                                          Commit.status != not_status).\
        order_by(desc(Commit.dt_commit)).\
        order_by(desc(Commit.id)).first()
    return commit


# Get the most recently built commit for project_name and commit_branch, we
# ignore commits with a status of "RETRY" as we want to retry these.
def getLastBuiltCommit(
        session, project_name, commit_branch, not_status="RETRY", type='rpm'):
    commit = session.query(Commit).filter(Commit.project_name == project_name,
                                          Commit.status != not_status,
                                          Commit.type == type,
                                          Commit.commit_branch ==
                                          commit_branch).\
        order_by(desc(Commit.dt_build)).\
        order_by(desc(Commit.id)).first()
    return commit


def getCommits(session, project=None, with_status=None, without_status=None,
               limit=1, order="desc", since=None, before=None, offset=0,
               component=None, type="rpm"):
    commits = session.query(Commit).filter(Commit.type == type)
    if project is not None:
        commits = commits.filter(Commit.project_name == project)
    if with_status is not None:
        commits = commits.filter(Commit.status == with_status)
    if without_status is not None:
        commits = commits.filter(Commit.status != without_status)
    if component is not None:
        commits = commits.filter(Commit.component == component)
    if since is not None:
        commits = commits.filter(Commit.dt_build > since)
    if before is not None:
        commits = commits.filter(or_(Commit.dt_build is None,
                                 Commit.dt_build < before))
    order_by = desc
    if order == "asc":
        order_by = asc
    commits = commits.order_by(order_by(Commit.id))
    if offset:
        commits = commits.offset(offset)
    if limit:
        commits = commits.limit(limit)
    return commits


@event.listens_for(Engine, "connect")
def connect(dbapi_connection, connection_record):
    connection_record.info['pid'] = os.getpid()


@event.listens_for(Engine, "checkout")
def checkout(dbapi_connection, connection_record, connection_proxy):
    pid = os.getpid()
    if connection_record.info['pid'] != pid:
        connection_record.connection = connection_proxy.connection = None
        raise exc.DisconnectionError(
            "Connection record belongs to pid %s, "
            "attempting to check out in pid %s" %
            (connection_record.info['pid'], pid)
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import jinja2
import logging
import os
import sh
import smtplib

from dlrn.config import getConfigOptions
from dlrn.reporting import get_commit_url
from email.mime.text import MIMEText

logger = logging.getLogger("dlrn-notifications")


def submit_review(commit, packages, env_vars):
    config_options = getConfigOptions()
    datadir = os.path.realpath(config_options.datadir)
    scriptsdir = os.path.realpath(config_options.scriptsdir)
    yumrepodir = os.path.join("repos", commit.getshardedcommitdir())

    project_name = commit.project_name

    for pkg in packages:
        if project_name == pkg['name']:
            break
    else:
        logger.error('Unable to find info for project'
                     ' %s' % project)
        return

    url = (get_commit_url(commit, pkg) + commit.commit_hash)
    env_vars.append('GERRIT_URL=%s' % url)
    env_vars.append('GERRIT_LOG=%s/%s' % (config_options.baseurl,
                                          commit.getshardedcommitdir()))
    maintainers = ','.join(pkg['maintainers'])
    env_vars.append('GERRIT_MAINTAINERS=%s' % maintainers)
    env_vars.append('GERRIT_TOPIC=%s' % config_options.gerrit_topic)
    logger.info('Creating a gerrit review using '
                'GERRIT_URL=%s '
                'GERRIT_MAINTAINERS=%s ' %
                (url, maintainers))

    run_cmd = []
    if env_vars:
        for env_var in env_vars:
            run_cmd.append(env_var)

    run_cmd.extend([os.path.join(scriptsdir, "submit_review.sh"),
                    project_name, os.path.join(datadir, yumrepodir),
                    datadir, config_options.baseurl,
                    os.path.realpath(commit.distgit_dir)])
    sh.env(run_cmd, _timeout=300)


def sendnotifymail(packages, commit):
    config_options = getConfigOptions()

    details = copy.copy(
        [package for package in packages
            if package["name"] == commit.project_name][0])

    email_to = details['maintainers']
    if not config_options.smtpserver:
        logger.info("Skipping notify email to %r" % email_to)
        return

    details["logurl"] = "%s/%s" % (config_options.baseurl,
                                   commit.getshardedcommitdir())
    # Render the notification template
    jinja_env = jinja2.Environment(
        loader=jinja2.FileSystemLoader([config_options.templatedir]))
    jinja_template = jinja_env.get_template("notification_email.j2")
    error_body = jinja_template.render(details=details)

    msg = MIMEText(error_body)
    msg['Subject'] = '[dlrn] %s master package build failed' % \
                     commit.project_name

    email_from = 'no-reply@delorean.com'
    msg['From'] = email_from
    msg['To'] = "packagers"

    logger.info("Sending notify email to %r" % email_to)
    try:
        s = smtplib.SMTP(config_options.smtpserver)
        s.sendmail(email_from, email_to, msg.as_string())
        s.quit()
    except smtplib.SMTPException as e:
        logger.error("An issue occured when sending"
                     "notify email to %r (%s)" % (email_to, e))
    finally:
        s.close()



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
import os
import shutil
import tempfile

from dlrn.config import ConfigOptions
from dlrn.drivers.coprdriver import CoprBuildDriver
from dlrn.tests import base
from six.moves import configparser


@mock.patch('sh.restorecon', create=True)
@mock.patch('sh.env', create=True)
class TestDriverCopr(base.TestCase):
    def setUp(self):
        super(TestDriverCopr, self).setUp()
        config = configparser.RawConfigParser()
        config.read("projects.ini")
        config.set("DEFAULT", "build_driver",
                   "dlrn.drivers.coprdriver.CoprBuildDriver")
        config.set('coprbuild_driver', 'coprid', 'foo/repo')
        self.config = ConfigOptions(config)
        self.temp_dir = tempfile.mkdtemp()
        # Create fake src.rpm
        with open('%s/pkg.src.rpm' % self.temp_dir, 'a') as fp:
            fp.write('')
        # Create fake build and download logs
        with open("%s/coprbuild.log" % self.temp_dir, 'a') as fp:
            fp.write("Created builds: 1234")
        with open('%s/coprdownload.log' % self.temp_dir, 'a') as fp:
            fp.write('')
        # Create fake download file structure
        os.mkdir(os.path.join(self.temp_dir, '1234'))
        target_chroot = os.path.join(
            self.temp_dir, '1234', 'fedora-rawhide-i386')
        os.mkdir(target_chroot)
        with open('%s/state.log.gz' % target_chroot, 'a') as fp:
            fp.write('')
        with open('%s/pkg.rpm' % target_chroot, 'a') as fp:
            fp.write('')

    def tearDown(self):
        super(TestDriverCopr, self).tearDown()
        shutil.rmtree(self.temp_dir)

    def test_build_package(self, env_mock, rc_mock):
        driver = CoprBuildDriver(cfg_options=self.config)
        driver.build_package(output_directory=self.temp_dir)

        expected = [mock.call(['copr', 'build',
                               self.config.coprid,
                               '%s/pkg.src.rpm' %
                               self.temp_dir],
                              _err=driver._process_copr_output,
                              _out=driver._process_copr_output,
                              _cwd=self.temp_dir,
                              _env={'PATH': '/usr/bin/'}),
                    mock.call(['copr', 'download-build', '-d',
                               '%s/1234' % self.temp_dir, '1234'],
                              _err=driver._process_copr_output,
                              _out=driver._process_copr_output,
                              _cwd=self.temp_dir,
                              _env={'PATH': '/usr/bin/'})]
        # 1- copr build (handled by env_mock)
        # 2- copr download-build (handled by env_mock)
        # 3- restorecon (handled by rc_mock)
        self.assertEqual(env_mock.call_count, 2)
        self.assertEqual(rc_mock.call_count, 1)
        self.assertEqual(env_mock.call_args_list, expected)

        # Make sure output_dir is as expected
        content = os.listdir(self.temp_dir)
        self.assertIn('coprdownload.log', content)
        self.assertIn('state.log.gz', content)
        self.assertIn('pkg.rpm', content)
        self.assertIn('pkg.src.rpm', content)
        self.assertNotIn('1234', content)

    def test_driver_config(self, env_mock, rc_mock):
        self.assertEqual(self.config.coprid, 'foo/repo')



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
import sh
from six.moves.urllib.request import urlopen
import sys
import tempfile

from dlrn.tests import base

from dlrn import db
from dlrn import remote
from dlrn import utils


def mocked_session(url):
    db_fd, filepath = tempfile.mkstemp()
    session = db.getSession("sqlite:///%s" % filepath)
    utils.loadYAML(session, './dlrn/tests/samples/commits_1.yaml')
    return session


def mocked_urlopen(url):
    if url.startswith('http://example.com'):
        fp = open('./dlrn/tests/samples/commits_remote.yaml', 'rb')
        return fp
    else:
        return urlopen(url)


@mock.patch('os.rename')
@mock.patch('os.symlink')
@mock.patch('dlrn.drivers.rdoinfo.RdoInfoDriver.getpackages')
@mock.patch.object(sh.Command, '__call__', autospec=True)
@mock.patch('dlrn.remote.post_build')
@mock.patch('dlrn.remote.getSession', side_effect=mocked_session)
@mock.patch('dlrn.remote.urlopen', side_effect=mocked_urlopen)
class TestRemote(base.TestCase):
    def test_remote(self, url_mock, db_mock, build_mock, sh_mock, gp_mock,
                    sl_mock, rn_mock):
        testargs = ["dlrn-remote", "--config-file", "projects.ini",
                    "--repo-url", "http://example.com/1/"]
        # There should be only one call to post_build(), even though there are
        # 2 commits in the commits_remote.yaml file. This is because the first
        # one is already in the database
        with mock.patch.object(sys, 'argv', testargs):
            remote.remote()
            build_mock.assert_called_once()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Add component to CIVote

Revision ID: 00a31f1f39c0
Revises: 7bed5ff86925
Create Date: 2019-10-07 10:47:29.032909

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import table, column


# revision identifiers, used by Alembic.
revision = '00a31f1f39c0'
down_revision = '7bed5ff86925'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('civotes', sa.Column('component', sa.String(64)))
    civotes = table('civotes',
                    column('id', sa.Integer),
                    column('component', sa.String))
    # For existing civotes, set component to None
    op.execute(civotes.update()
               .where(civotes.c.id == civotes.c.id)
               .values(component=None))


def downgrade():
    with op.batch_alter_table('civotes') as batch_op:
        batch_op.drop_column('component')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from datetime import datetime
from functools import partial
from operator import itemgetter
import os
import shutil
from time import gmtime
from time import strftime

from six.moves.urllib import parse

import jinja2

from dlrn.config import getConfigOptions
from dlrn.db import Commit
from dlrn.db import getCommits


def get_commit_url(commit, pkg):
    try:
        upstream_url = parse.urlsplit(pkg["upstream"])

        if upstream_url.netloc == "git.openstack.org":
            commit_url = ("http",
                          upstream_url.netloc,
                          "/cgit%s/commit/?id=" % upstream_url.path,
                          "", "", "")
            commit_url = parse.urlunparse(commit_url)
        elif upstream_url.netloc == "github.com":
            commit_url = ("https",
                          upstream_url.netloc,
                          "%s/commit/" % upstream_url.path,
                          "", "", "")
            commit_url = parse.urlunparse(commit_url)
        else:
            # Fallback when no cgit URL can be defined
            commit_url = pkg["upstream"]
    except KeyError:
        # This should not happen, but pkg['upstream'] may not be present
        # after some error in the gitrepo driver
        commit_url = ''

    return commit_url


def _jinja2_filter_strftime(date, fmt="%Y-%m-%d %H:%M:%S"):
    gmdate = gmtime(date)
    return "%s" % strftime(fmt, gmdate)


def _jinja2_filter_get_commit_url(commit, packages):
    for pkg in packages:
        project = pkg["name"]
        if project == commit.project_name:
            return get_commit_url(commit, pkg)
    return "???"


def genreports(packages, head_only, session, all_commits):
    config_options = getConfigOptions()

    # Generate report of the last 300 package builds
    target = config_options.target
    src = config_options.source
    reponame = config_options.reponame
    templatedir = config_options.templatedir
    project_name = config_options.project_name
    datadir = config_options.datadir
    repodir = os.path.join(datadir, "repos")

    css_file = os.path.join(templatedir, 'stylesheets/styles.css')

    # create directories
    if not os.path.exists(repodir):
        os.makedirs(repodir)

    # configure jinja and filters
    jinja_env = jinja2.Environment(
        loader=jinja2.FileSystemLoader([templatedir]))
    jinja_env.filters["strftime"] = _jinja2_filter_strftime
    jinja_env.filters["get_commit_url"] = \
        partial(_jinja2_filter_get_commit_url, packages=packages)

    # generate build report
    commits = getCommits(session, without_status="RETRY", limit=300)
    jinja_template = jinja_env.get_template("report.j2")
    content = jinja_template.render(reponame=reponame,
                                    src=src,
                                    project_name=project_name,
                                    target=target,
                                    commits=commits)
    shutil.copy2(css_file, os.path.join(repodir, "styles.css"))
    report_file = os.path.join(repodir, "report.html")
    with open(report_file, "w") as fp:
        fp.write(content)

    # Generate status report
    if head_only:
        msg = "(all commit not built)"
    else:
        msg = ""

    pkgs = []
    # Find the most recent successfull build
    # then report on failures since then
    for package in packages:
        name = package["name"]
        commits = getCommits(session, project=name, limit=1)

        # No builds
        if commits.count() == 0:
            continue

        pkgs.append(package)
        last_build = commits.first()
        package["last_build"] = last_build

        # last build was successul
        if last_build.status == "SUCCESS":
            continue

        # Retrieve last successful build
        commits = getCommits(session, project=name, with_status="SUCCESS",
                             limit=1)

        # No successful builds
        if commits.count() == 0:
            commits = getCommits(session, project=name, with_status="FAILED",
                                 order="asc")
            package["first_failure"] = commits.first()
            package["days"] = -1
            continue

        last_success = commits.first()
        last_success_dt = last_success.dt_build

        commits = getCommits(session, project=name, with_status="FAILED",
                             order="asc", limit=None)
        commits = commits.filter(Commit.dt_build > last_success_dt)
        package["first_failure"] = commits.first()
        package["days"] = (datetime.now() -
                           datetime.fromtimestamp(last_success_dt)).days

    pkgs = sorted(pkgs, key=itemgetter("name"))
    jinja_template = jinja_env.get_template("status_report.j2")
    content = jinja_template.render(msg=msg,
                                    reponame=reponame,
                                    src=src,
                                    project_name=project_name,
                                    target=target,
                                    pkgs=pkgs)

    report_file = os.path.join(repodir, "status_report.html")
    with open(report_file, "w") as fp:
        fp.write(content)

    jinja_template = jinja_env.get_template("status_report_csv.j2")
    content = jinja_template.render(msg=msg,
                                    reponame=reponame,
                                    src=src,
                                    project_name=project_name,
                                    target=target,
                                    pkgs=pkgs)

    report_file = os.path.join(repodir, "status_report.csv")
    with open(report_file, "w") as fp:
        fp.write(content)

    # Create a report for the pending packages
    jinja_template = jinja_env.get_template("queue.j2")
    pending_commits = []
    for commit in all_commits:
        old_commit = getCommits(session, project=commit.project_name,
                                without_status="RETRY", limit=None).filter(
            Commit.commit_hash == commit.commit_hash).filter(
            Commit.distro_hash == commit.distro_hash).first()
        if not old_commit:
            pending_commits.append(commit)

    content = jinja_template.render(reponame=reponame,
                                    src=src,
                                    target=target,
                                    commits=pending_commits)
    report_file = os.path.join(repodir, "queue.html")
    with open(report_file, "w") as fp:
        fp.write(content)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""change user/usernames column length

Revision ID: 2a0313a8a7d6
Revises: cab7697f6564
Create Date: 2018-02-27 15:35:04.667089

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '2a0313a8a7d6'
down_revision = 'ade85b2396bc'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("civotes") as batch_op:
        batch_op.drop_constraint('civ_user_fk', type_='foreignkey')

    with op.batch_alter_table("promotions") as batch_op:
        batch_op.drop_constraint('prom_user_fk', type_='foreignkey')

    with op.batch_alter_table("users") as batch_op:
        batch_op.alter_column('username', existing_type=sa.String(256),
                              type_=sa.String(255))

    with op.batch_alter_table("civotes") as batch_op:
        batch_op.alter_column('user', existing_type=sa.String(256),
                              type_=sa.String(255))
        batch_op.create_foreign_key(
            constraint_name="civ_user_fk",
            referent_table="users",
            local_cols=["user"],
            remote_cols=["username"])

    with op.batch_alter_table("promotions") as batch_op:
        batch_op.alter_column('user', existing_type=sa.String(256),
                              type_=sa.String(255))
        batch_op.create_foreign_key(
            constraint_name="prom_user_fk",
            referent_table="users",
            local_cols=["user"],
            remote_cols=["username"])


def downgrade():
    with op.batch_alter_table("civotes") as batch_op:
        batch_op.drop_constraint('civ_user_fk', type_='foreignkey')

    with op.batch_alter_table("promotions") as batch_op:
        batch_op.drop_constraint('prom_user_fk', type_='foreignkey')

    with op.batch_alter_table("users") as batch_op:
        batch_op.alter_column('username', existing_type=sa.String(255),
                              type_=sa.String(256))

    with op.batch_alter_table("civotes") as batch_op:
        batch_op.alter_column('user', existing_type=sa.String(255),
                              type_=sa.String(256))
        batch_op.create_foreign_key(
            constraint_name="civ_user_fk",
            referent_table="users",
            local_cols=["user"],
            remote_cols=["username"])

    with op.batch_alter_table("promotions") as batch_op:
        batch_op.alter_column('user', existing_type=sa.String(255),
                              type_=sa.String(256))
        batch_op.create_foreign_key(
            constraint_name="prom_user_fk",
            referent_table="users",
            local_cols=["user"],
            remote_cols=["username"])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# The GitRepoDriver provides the following:
#
# 1- A getpackages function based on finding all directories inside a specific
#    git repo, where each directory represents a package
#
# 2- A getinfo function based on a single-distgit repo paradigm
#
# 3- A preprocess function using renderspec on any *.spec.j2 file found in the
#    distgit.

import logging
import os
import re
import sh

from dlrn.db import Commit
from dlrn.drivers.pkginfo import PkgInfoDriver
from dlrn.repositories import getsourcebranch
from dlrn.repositories import refreshrepo
from dlrn.utils import run_external_preprocess
from pymod2pkg import module2package
from pymod2pkg import module2upstream
from rdopkg.utils import specfile
from six.moves import urllib
from six.moves.urllib.request import urlopen

logger = logging.getLogger("dlrn-gitrepo-driver")

version_match = re.compile(r'\W*set upstream_version\D+([\w.]+).*')
wrong_match = re.compile(r'\W*set upstream_version\D+\(\).*')

base_urls = ['https://opendev.org/openstack',
             'https://opendev.org/x',
             'https://opendev.org/opendev']


def check_url(url):
    logger.info("Checking url %s" % url)
    try:
        urlopen(url)
        # URL found
        return True
    except (urllib.error.HTTPError, urllib.error.URLError):
        # Trouble finding URL
        return False


class GitRepoDriver(PkgInfoDriver):
    DRIVER_CONFIG = {
        'gitrepo_driver': {
            'gitrepo_repo': {'name': 'repo'},
            'gitrepo_dirs': {'name': 'directory', 'type': 'list'},
            'skip_dirs': {'name': 'skip', 'type': 'list'},
            'use_version_from_spec': {'type': 'boolean'},
            'keep_tarball': {'type': 'boolean'},
        }
    }

    def __init__(self, *args, **kwargs):
        super(GitRepoDriver, self).__init__(*args, **kwargs)

    def _get_version_from_pkg(self, packagepath, package):
        self.preprocess(package_name=package)
        pkgdir = os.path.join(packagepath, package)
        for pkgfile in os.listdir(pkgdir):
            if pkgfile.endswith('.spec'):
                spec = specfile.Spec(fn=os.path.join(packagepath, package,
                                                     pkgfile))
                version = spec.get_tag('Version')
                release = spec.get_tag('Release')

        if release.startswith('0'):
            # This is a pre-release version, so we are following trunk
            return None
        else:
            # This is a tagged release
            return version

    def getpackages(self, **kwargs):
        repo = self.config_options.gitrepo_repo
        distro_branch = self.config_options.distro
        datadir = self.config_options.datadir
        skip_dirs = self.config_options.skip_dirs
        dev_mode = kwargs.get('dev_mode')
        packages = []

        gitpath = os.path.join(datadir, 'package_info')
        if not os.path.exists(gitpath):
            sh.git.clone(repo, gitpath)

        if not dev_mode:
            git = sh.git.bake(_cwd=gitpath, _tty_out=False, _timeout=3600)
            git.fetch("origin")
            # Use the the configured distro branch, fall back to master
            # if it fails
            try:
                git.reset("--hard", "origin/%s" % distro_branch)
            except Exception:
                logger.info("Falling back to master")
                git.reset("--hard", "origin/master")

        for basepath in self.config_options.gitrepo_dirs:
            path = basepath.strip('/')
            packagepath = os.path.join(gitpath, path)

            for package in os.listdir(packagepath):
                if (os.path.isdir(os.path.join(packagepath, package)) and
                   package not in skip_dirs):
                    pkg_hash = {}
                    pkg_hash['name'] = package
                    pkg_hash['maintainers'] = 'test@example.com'
                    pkg_hash['master-distgit'] = (repo + '/' + path + '/' +
                                                  package)
                    pkg_hash['upstream'] = 'Unknown'
                    if self.config_options.use_version_from_spec is True:
                        version = None
                        # Try to deduce version from spec template
                        pkgdir = os.path.join(packagepath, package)
                        for pkgfile in os.listdir(pkgdir):
                            if pkgfile.endswith('.j2'):
                                with open(os.path.join(pkgdir, pkgfile)) as fp:
                                    j2content = fp.readlines()
                                for line in j2content:
                                    # Make sure we are not matching the wrong
                                    # version of the upstream_version() macro
                                    m = wrong_match.match(line)
                                    if m is not None:
                                        # We are deducing the version using
                                        # the source tarball. This is a bit
                                        # mroe complex, since we need to run
                                        # renderspec and find the version in
                                        # the resulting spec
                                        version = self._get_version_from_pkg(
                                            packagepath, package)
                                        logger.info(
                                            "Got version %s for %s from the "
                                            "spec" % (version, package))
                                        break
                                    # Does template define upstream_version?
                                    m = version_match.match(line)
                                    if m is not None:
                                        version = m.group(1)
                                        break
                                    # Otherwise, we're using a direct version
                                    if line.startswith('Version:'):
                                        version = line.split(':')[1].strip().\
                                            replace('~', '')
                                        break

                        if version is not None:
                            pkg_hash['source-branch'] = version
                    packages.append(pkg_hash)
        return packages

    def getinfo(self, **kwargs):
        project = kwargs.get('project')
        package = kwargs.get('package')
        since = kwargs.get('since')
        local = kwargs.get('local')
        dev_mode = kwargs.get('dev_mode')
        datadir = self.config_options.datadir

        for url in base_urls:
            if check_url("%s/%s" % (url, module2upstream(package['name']))):
                package['upstream'] = ("%s/%s" %
                                       (url,
                                        module2upstream(package['name'])))
                break
        else:
            logger.error("Could not find upstream URL for project %s" %
                         package)

        repo = package['upstream']

        distro_dir = self.distgit_dir(package['name'])
        source_branch = getsourcebranch(package)

        if dev_mode is True:
            distro_hash = "dev"
            dt_distro = 0  # Doesn't get used in dev mode
        else:
            # Get distro_hash from last commit in distgit directory
            git = sh.git.bake(_cwd=distro_dir, _tty_out=False)
            repoinfo = str(git.log("--pretty=format:%H %ct", "-1", ".")
                           ).strip().split(" ")
            distro_hash = repoinfo[0]
            dt_distro = repoinfo[1]

        # repo is usually a string, but if it contains more then one entry we
        # git clone into a project subdirectory
        repos = [repo]
        if isinstance(repo, list):
            repos = repo
        project_toprocess = []
        for repo in repos:
            repo_dir = os.path.join(datadir, project)
            if len(repos) > 1:
                repo_dir = os.path.join(repo_dir, os.path.split(repo)[1])
            try:
                source_branch, _, _ = refreshrepo(repo, repo_dir,
                                                  source_branch, local=local)
            except Exception:
                # The error was already logged by refreshrepo, and the only
                # side-effect is that we are not adding this commit to the
                # list of commits to be processed, so we can ignore it and
                # move on to the next repo
                continue

            git = sh.git.bake(_cwd=repo_dir, _tty_out=False)
            # Git gives us commits already sorted in the right order
            lines = git.log("--pretty=format:'%ct %H'", since,
                            "--first-parent", "--reverse")

            for line in lines:
                dt, commit_hash = str(line).strip().strip("'").split(" ")
                commit = Commit(dt_commit=float(dt), project_name=project,
                                type='rpm',
                                commit_hash=commit_hash, repo_dir=repo_dir,
                                distro_hash=distro_hash, dt_distro=dt_distro,
                                distgit_dir=distro_dir,
                                commit_branch=source_branch,
                                dt_extended=0, extended_hash=None,
                                component=None)
                project_toprocess.append(commit)

        return project_toprocess

    def preprocess(self, **kwargs):
        package_name = kwargs.get('package_name')
        commit_hash = kwargs.get('commit_hash')
        distgit_dir = self.distgit_dir(package_name)
        output_filename = "%s.spec" % module2package(package_name, 'fedora')
        source_dir = "%s/%s" % (self.config_options.datadir, package_name)
        preprocess = sh.renderspec.bake(_cwd=distgit_dir,
                                        _tty_out=False, _timeout=3600)
        preprocess('--spec-style', 'fedora', '--epoch',
                   '../../epoch/fedora.yaml', '--output', output_filename)
        # Replace %{version} with %{upstream_version} in spec
        # This is required by rpm-packaging specs
        for specf in os.listdir(distgit_dir):
            if specf.endswith(".spec"):
                filename = os.path.join(distgit_dir, specf)
                with open(filename, 'r+') as fp:
                    spec = fp.read()
                    spec = re.sub(r'-%{version}', '-%{upstream_version}', spec)
                    fp.seek(0)
                    fp.write(spec)

        for custom_preprocess in self.config_options.custom_preprocess:
            if custom_preprocess != '':
                run_external_preprocess(
                    cmdline=custom_preprocess,
                    pkgname=package_name,
                    distgit=distgit_dir,
                    source_dir=source_dir,
                    commit_hash=commit_hash)

    def distgit_dir(self, package_name):
        datadir = self.config_options.datadir
        # Check all potential base directories
        for basepath in self.config_options.gitrepo_dirs:
            path = basepath.strip('/')
            fullpath = os.path.join(datadir, 'package_info', path,
                                    package_name)
            if os.path.exists(fullpath):
                return fullpath



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# PkgInfoDriver derived classes expose the following functions:
#
# getpackages(). This function will return an array of hashes. Each individual
# hash must contain the following mandatory parameters (others are optional):
# - 'name' : package name
# - 'upstream': URL for upstream repo
# - 'master-distgit': URL for distgit repo
# - 'maintainers': list of e-mail addresses for package maintainers
#
# getinfo(). This function will return a list of commits to be processed for a
#            specific package.
#
# preprocess(). This function will run any required pre-processing for the spec
#               files.
#
# distgit_dir(). This function will return the distgit repo directory for a
#                given package name.


class PkgInfoDriver(object):
    def __init__(self, *args, **kwargs):
        self.packages = []
        self.config_options = kwargs.get('cfg_options')

    def getpackages(self):
        return self.packages

    def getinfo(self):
        return None

    def preprocess(self):
        return

    def distgit_dir(self, package_name):
        return ''



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
import os
import passlib.hash
import sys
import tempfile

from dlrn import db
from dlrn.tests import base
from dlrn import user
from dlrn import utils


class TestUser(base.TestCase):
    def setUp(self):
        super(TestUser, self).setUp()
        self.db_fd, filepath = tempfile.mkstemp()
        self.session = db.getSession("sqlite:///%s" % filepath)
        utils.loadYAML(self.session, './dlrn/tests/samples/commits_2.yaml')

    def tearDown(self):
        super(TestUser, self).tearDown()
        os.close(self.db_fd)

    def mocked_session(url):
        return self.session

    @mock.patch('dlrn.user.getSession')
    def test_user_create(self, db_mock):
        db_mock.return_value = self.session
        testargs = ["dlrn-user", "create", "--username",
                    "user", "--password", "password"]
        with mock.patch.object(sys, 'argv', testargs):
            user.user_manager()

        # Now check that the user was created
        myuser = self.session.query(db.User).filter(
            db.User.username == 'user').first()
        self.assertEqual(myuser.username, 'user')

    @mock.patch('dlrn.user.getSession')
    def test_user_create_duplicate(self, db_mock):
        db_mock.return_value = self.session
        testargs = ["dlrn-user", "create", "--username",
                    "user", "--password", "password"]
        with mock.patch.object(sys, 'argv', testargs):
            user.user_manager()
            return_value = user.user_manager()

        # Check that we got a failure
        self.assertEqual(return_value, -1)

    @mock.patch('dlrn.user.getpass')
    @mock.patch('dlrn.user.getSession')
    def test_user_create_interactive(self, db_mock, gp_mock):
        db_mock.return_value = self.session
        gp_mock.return_value = 'password'
        testargs = ["dlrn-user", "create", "--username", "user"]
        with mock.patch.object(sys, 'argv', testargs):
            user.user_manager()

        # Now check that the user was created
        myuser = self.session.query(db.User).filter(
            db.User.username == 'user').first()
        self.assertEqual(myuser.username, 'user')

    @mock.patch('dlrn.user.getSession')
    def test_user_update(self, db_mock):
        db_mock.return_value = self.session
        testargs = ["dlrn-user", "update", "--username",
                    "foo", "--password", "password"]
        with mock.patch.object(sys, 'argv', testargs):
            user.user_manager()

        # Now check that the user was properly updated
        myuser = self.session.query(db.User).filter(
            db.User.username == 'foo').first()
        self.assertEqual(
            passlib.hash.sha512_crypt.verify('password', myuser.password),
            True)

    @mock.patch('dlrn.user.getSession')
    def test_user_update_non_existing(self, db_mock):
        db_mock.return_value = self.session
        testargs = ["dlrn-user", "update", "--username", "noone",
                    "--password", "foo"]
        with mock.patch.object(sys, 'argv', testargs):
            return_value = user.user_manager()

        # Check that we got a failure
        self.assertEqual(return_value, -1)

    @mock.patch('dlrn.user.getSession')
    def test_user_delete_force(self, db_mock):
        db_mock.return_value = self.session
        testargs = ["dlrn-user", "create", "--username",
                    "user2", "--password", "password"]
        with mock.patch.object(sys, 'argv', testargs):
            user.user_manager()

        testargs = ["dlrn-user", "delete", "--username",
                    "user2", "--force"]
        with mock.patch.object(sys, 'argv', testargs):
            user.user_manager()

        # Now check that the user was deleted
        myuser = self.session.query(db.User).filter(
            db.User.username == 'user2').first()
        self.assertEqual(myuser, None)

    @mock.patch('dlrn.user.input')
    @mock.patch('dlrn.user.getSession')
    def test_user_delete_confirm(self, db_mock, in_mock):
        db_mock.return_value = self.session
        in_mock.return_value = 'YES'
        testargs = ["dlrn-user", "create", "--username",
                    "user3", "--password", "password"]
        with mock.patch.object(sys, 'argv', testargs):
            user.user_manager()

        testargs = ["dlrn-user", "delete", "--username", "user3"]
        with mock.patch.object(sys, 'argv', testargs):
            user.user_manager()

        # Now check that the user was deleted
        myuser = self.session.query(db.User).filter(
            db.User.username == 'user3').first()
        self.assertEqual(myuser, None)

    @mock.patch('dlrn.user.input')
    @mock.patch('dlrn.user.getSession')
    def test_user_delete_without_confirm(self, db_mock, in_mock):
        db_mock.return_value = self.session
        in_mock.return_value = 'abc'
        testargs = ["dlrn-user", "create", "--username",
                    "user4", "--password", "password"]
        with mock.patch.object(sys, 'argv', testargs):
            user.user_manager()

        testargs = ["dlrn-user", "delete", "--username", "user4"]
        with mock.patch.object(sys, 'argv', testargs):
            user.user_manager()

        # Now check that the user was deleted
        myuser = self.session.query(db.User).filter(
            db.User.username == 'user4').first()
        self.assertNotEqual(myuser, None)

    @mock.patch('dlrn.user.getSession')
    def test_user_delete_non_existing(self, db_mock):
        db_mock.return_value = self.session
        testargs = ["dlrn-user", "delete", "--username", "noone"]
        with mock.patch.object(sys, 'argv', testargs):
            return_value = user.user_manager()

        # Check that we got a failure
        self.assertEqual(return_value, -1)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# This utility script will recreate all promotion symlinks from the database,
# using the latest promotion for each.

import argparse
import os
import sys

from dlrn.db import Commit
from dlrn.db import getSession
from dlrn.db import Promotion
from six.moves import configparser
from sqlalchemy import desc


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--config-file',
                        help="Config file (required)", required=True)
    parser.add_argument('--noop',
                        help="Preview actions but do not execute them",
                        action="store_true")

    options = parser.parse_args(sys.argv[1:])

    cp = configparser.RawConfigParser()
    cp.read(options.config_file)
    datadir = os.path.realpath(cp.get('DEFAULT', 'datadir'))
    session = getSession(cp.get('DEFAULT', 'database_connection'))

    # We need to find promotions, and for each promotion name create the
    # corresponding symlink
    promotions = session.query(Promotion.promotion_name).distinct()

    promotion_list = []
    for promotion in promotions:
        promotion_list.append(promotion.promotion_name)

    # Find latest promotions for each promotion name, and re-do the symlinks
    for name in promotion_list:
        promotion = session.query(Promotion).\
            order_by(desc(Promotion.timestamp)).\
            filter(Promotion.promotion_name == name).first()

        commit = session.query(Commit).\
            filter(Commit.id == promotion.commit_id).first()

        repo_dir = os.path.join(datadir, "repos", commit.getshardedcommitdir())
        symlink_path = os.path.join(datadir, "repos", name)
        print("Going to symlink %s to %s" % (symlink_path, repo_dir))

        if not options.noop:
            try:
                os.symlink(os.path.relpath(repo_dir,
                                           os.path.join(datadir, "repos")),
                           symlink_path + "_")
                os.rename(symlink_path + "_", symlink_path)
            except Exception as e:
                print("Symlink creation failed: %s", e)


if __name__ == '__main__':
    main()
    exit(0)



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import mock
import sh

from dlrn.config import ConfigOptions
from dlrn import repositories
from dlrn.tests import base

from six.moves import configparser


def _aux_sh(*args):
    call = args[0]
    if call == '-f':
        raise sh.ErrorReturnCode_1('blabla'.encode(), ''.encode(),
                                   ''.encode())
    return


@mock.patch.object(sh.Command, '__call__', autospec=True)
class TestRefreshRepo(base.TestCase):

    def test_clone_if_not_cloned(self, sh_mock):
        repositories.refreshrepo('url', 'path', branch='branch')
        expected = [mock.call(sh.git.clone, 'url', 'path'),
                    mock.call(sh.git.fetch, 'origin'),
                    mock.call(sh.git.checkout, '-f', 'branch'),
                    mock.call(sh.git.reset, '--hard', 'origin/branch'),
                    mock.call(sh.git.log, '--pretty=format:%H %ct', '-1', '.')]
        self.assertEqual(sh_mock.call_args_list, expected)

    @mock.patch('os.path.exists', return_value=True)
    @mock.patch('shutil.rmtree', return_value=True)
    def test_dont_clone_if_cloned(self, path_mock, shutil_mock, sh_mock):
        repositories.refreshrepo('url', 'path', branch='branch')
        expected = [mock.call(sh.git, 'remote', '-v'),
                    mock.call(sh.git.clone, 'url', 'path'),
                    mock.call(sh.git.fetch, 'origin'),
                    mock.call(sh.git.checkout, '-f', 'branch'),
                    mock.call(sh.git.reset, '--hard', 'origin/branch'),
                    mock.call(sh.git.log, '--pretty=format:%H %ct', '-1', '.')]
        self.assertEqual(sh_mock.call_args_list, expected)

    @mock.patch('os.path.exists', return_value=True)
    def test_dont_fetch_if_local_repo_exists(self, path_mock, sh_mock):
        repositories.refreshrepo('url', 'path', branch='branch', local=True)
        expected = [mock.call(sh.git.log, '--pretty=format:%H %ct', '-1', '.')]
        self.assertEqual(sh_mock.call_args_list, expected)

    def test_clone_fetch_if_local_repo_missing(self, sh_mock):
        repositories.refreshrepo('url', 'path', branch='branch', local=True)
        expected = [mock.call(sh.git.clone, 'url', 'path'),
                    mock.call(sh.git.fetch, 'origin'),
                    mock.call(sh.git.checkout, '-f', 'branch'),
                    mock.call(sh.git.reset, '--hard', 'origin/branch'),
                    mock.call(sh.git.log, '--pretty=format:%H %ct', '-1', '.')]
        self.assertEqual(sh_mock.call_args_list, expected)

    def test_clone_no_fallback(self, sh_mock):
        config = configparser.RawConfigParser()
        config.read("projects.ini")
        config.set('DEFAULT', 'fallback_to_master', '0')
        self.config = ConfigOptions(config)
        # We need to redefine the mock object again, to use a side effect
        # that will fail in the git checkout call. A bit convoluted, but
        # it works
        with mock.patch.object(sh.Command, '__call__') as new_mock:
            new_mock.side_effect = _aux_sh
            self.assertRaises(sh.ErrorReturnCode_1, repositories.refreshrepo,
                              'url', 'path', branch='branch')
            expected = [mock.call('url', 'path'),
                        mock.call('origin'),
                        mock.call('-f', 'branch')]
            self.assertEqual(new_mock.call_args_list, expected)

    def test_clone_no_fallback_default(self, sh_mock):
        config = configparser.RawConfigParser()
        config.read("projects.ini")
        config.set('DEFAULT', 'fallback_to_master', '1')
        self.config = ConfigOptions(config)
        with mock.patch.object(sh.Command, '__call__') as new_mock:
            new_mock.side_effect = _aux_sh
            self.assertRaises(sh.ErrorReturnCode_1, repositories.refreshrepo,
                              'url', 'path', branch='rpm-master')
            expected = [mock.call('url', 'path'),
                        mock.call('origin'),
                        mock.call('-f', 'rpm-master')]
            self.assertEqual(new_mock.call_args_list, expected)

    def test_clone_no_fallback_var(self, sh_mock):
        config = configparser.RawConfigParser()
        config.read("projects.ini")
        config.set('DEFAULT', 'fallback_to_master', '1')
        config.set('DEFAULT', 'nonfallback_branches', '^foo-')
        self.config = ConfigOptions(config)
        with mock.patch.object(sh.Command, '__call__') as new_mock:
            new_mock.side_effect = _aux_sh
            self.assertRaises(sh.ErrorReturnCode_1, repositories.refreshrepo,
                              'url', 'path', branch='foo-bar')
            expected = [mock.call('url', 'path'),
                        mock.call('origin'),
                        mock.call('-f', 'foo-bar')]
            self.assertEqual(new_mock.call_args_list, expected)

    def test_clone_fallback_var(self, sh_mock):
        config = configparser.RawConfigParser()
        config.read("projects.ini")
        config.set('DEFAULT', 'fallback_to_master', '1')
        config.set('DEFAULT', 'nonfallback_branches', '^foo-')
        self.config = ConfigOptions(config)
        with mock.patch.object(sh.Command, '__call__') as new_mock:
            new_mock.side_effect = _aux_sh
            result = repositories.refreshrepo('url', 'path', branch='bar')
            self.assertEqual(result, ['master', 'None'])
            expected = [mock.call('url', 'path'),
                        mock.call('origin'),
                        mock.call('-f', 'bar'),
                        mock.call('master'),
                        mock.call('--hard', 'origin/master'),
                        mock.call('--pretty=format:%H %ct', '-1', '.')]
            self.assertEqual(new_mock.call_args_list, expected)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import fcntl
import hashlib
import logging
import os
import re
import sh
import sys
import yaml

import sqlalchemy

from contextlib import contextmanager
from dlrn.db import CIVote
from dlrn.db import CIVote_Aggregate
from dlrn.db import Commit
from dlrn.db import getSession
from dlrn.db import Project
from dlrn.db import Promotion
from dlrn.db import User

re_known_errors = re.compile('Error: Nothing to do|'
                             'Error downloading packages|'
                             'No more mirrors to try|'
                             'Cannot retrieve metalink for repository|'
                             'Could not retrieve mirrorlist|'
                             'Failed to synchronize cache for repo|'
                             'No route to host|'
                             'Device or resource busy|'
                             'Could not resolve host|'
                             'Temporary failure in name resolution|'
                             'distroinfo.exception.CommandFailed: Command '
                             'failed with return code 128: git|'
                             'Error fetching remote|'
                             'Connection timed out')


logger = logging.getLogger("dlrn-utils")


# Import a Python class
def import_class(import_str):
    mod_str, _sep, class_str = import_str.rpartition('.')
    __import__(mod_str)
    try:
        myclass = getattr(sys.modules[mod_str], class_str)
        return myclass
    except AttributeError:
        raise ImportError('Cannot find class %s' % class_str)


# Import a Python object
def import_object(import_str, *args, **kwargs):
    myclass = import_class(import_str)
    return myclass(*args, **kwargs)


# Load a yaml file into a db session, used to populate a in memory database
# during tests
def loadYAML(session, yamlfile):
    with open(yamlfile) as fp:
        data = yaml.safe_load(fp)

    for commit in data['commits']:
        c = Commit(**commit)
        # We need a special case for extended_hash, which could be "None"
        if c.extended_hash == 'None':
            c.extended_hash = None
        # Retro compatibility before commit type
        if not c.type:
            c.type = "rpm"
        session.add(c)
        session.commit()
    try:
        for project in data['projects']:
            p = Project(**project)
            session.add(p)
            session.commit()
    except KeyError:
        pass   # No projects in yaml, just ignore
    try:
        for civote in data['civotes']:
            vote = CIVote(**civote)
            session.add(vote)
            session.commit()
    except KeyError:
        pass   # No civotes in yaml, just ignore
    try:
        for civote in data['civotes_agg']:
            vote = CIVote_Aggregate(**civote)
            session.add(vote)
            session.commit()
    except KeyError:
        pass   # No civotes_agg in yaml, just ignore
    try:
        for user in data['users']:
            my_user = User(**user)
            session.add(my_user)
            session.commit()
    except KeyError:
        pass   # No users in yaml, just ignore
    try:
        for promotion in data['promotions']:
            p = Promotion(**promotion)
            session.add(p)
            session.commit()
    except KeyError:
        pass   # No promotions in yaml, just ignore

    session.commit()


# Load a yaml file into a list of commits
def loadYAML_list(yamlfile):
    with open(yamlfile) as fp:
        data = yaml.safe_load(fp)

    commit_list = []
    for commit in data['commits']:
        c = Commit(**commit)
        # We need a special case for extended_hash, which could be "None"
        if c.extended_hash == 'None':
            c.extended_hash = None
        # Retro compatibility before commit type
        if not c.type:
            c.type = "rpm"
        commit_list.append(c)

    return commit_list


# Save a database to yaml, this is a helper function to assist in creating
# yaml files for unit tests.
def saveYAML(session, yamlfile):
    data = {}

    attrs = []
    for a in dir(Commit):
        if type(getattr(Commit, a)) == \
                sqlalchemy.orm.attributes.InstrumentedAttribute:
            attrs.append(a)
    data['commits'] = []
    for commit in session.query(Commit).all():
        d = {}
        for a in attrs:
            d[a] = str(getattr(commit, a))
        data['commits'].append(d)

    attrs = []
    for a in dir(Project):
        if type(getattr(Project, a)) == \
                sqlalchemy.orm.attributes.InstrumentedAttribute:
            attrs.append(a)
    data['projects'] = []
    for project in session.query(Project).all():
        d = {}
        for a in attrs:
            d[a] = str(getattr(project, a))
        data['projects'].append(d)

    attrs = []
    for a in dir(CIVote):
        if type(getattr(CIVote, a)) == \
                sqlalchemy.orm.attributes.InstrumentedAttribute:
            attrs.append(a)
    data['civotes'] = []
    for vote in session.query(CIVote).all():
        d = {}
        for a in attrs:
            d[a] = str(getattr(vote, a))
        data['civotes'].append(d)

    attrs = []
    for a in dir(CIVote_Aggregate):
        if type(getattr(CIVote_Aggregate, a)) == \
                sqlalchemy.orm.attributes.InstrumentedAttribute:
            attrs.append(a)
    data['civotes_agg'] = []
    for vote in session.query(CIVote_Aggregate).all():
        d = {}
        for a in attrs:
            d[a] = str(getattr(vote, a))
        data['civotes_agg'].append(d)

    attrs = []
    for a in dir(User):
        if type(getattr(User, a)) == \
                sqlalchemy.orm.attributes.InstrumentedAttribute:
            attrs.append(a)
    data['users'] = []
    for user in session.query(User).all():
        d = {}
        for a in attrs:
            d[a] = str(getattr(user, a))
        data['users'].append(d)

    attrs = []
    for a in dir(Promotion):
        if type(getattr(Promotion, a)) == \
                sqlalchemy.orm.attributes.InstrumentedAttribute:
            attrs.append(a)
    data['promotions'] = []
    for promotion in session.query(Promotion).all():
        d = {}
        for a in attrs:
            d[a] = str(getattr(promotion, a))
        data['promotions'].append(d)

    with open(yamlfile, 'w') as fp:
        fp.write(yaml.dump(data, default_flow_style=False))


# Save a single commit to yaml
def saveYAML_commit(commit, yamlfile):
    data = {}
    attrs = []
    for a in dir(Commit):
        if type(getattr(Commit, a)) == \
                sqlalchemy.orm.attributes.InstrumentedAttribute:
            attrs.append(a)
    data['commits'] = []
    # Add commit
    d = {}
    for a in attrs:
        d[a] = str(getattr(commit, a))
    data['commits'].append(d)
    with open(yamlfile, 'w') as fp:
        fp.write(yaml.dump(data, default_flow_style=False))


def dumpshas2file(shafile, commit, source_repo, distgit_repo,
                  status, timestamp, component, rpmlist):
    shafile.write("%s,%s,%s,%s,%s,%s,%d,%s,%s,%s\n" % (commit.project_name,
                                                       source_repo,
                                                       commit.commit_hash,
                                                       distgit_repo,
                                                       commit.distro_hash,
                                                       status,
                                                       timestamp,
                                                       component,
                                                       commit.extended_hash,
                                                       getNVRfromlist(rpmlist))
                  )


def getNVRfromlist(rpmlist):
    # Return a string with the source package NVR
    for pkg in rpmlist:
        if pkg.endswith(".src.rpm"):
            return pkg.split('/')[-1].split('.src.rpm')[0]
    return ""


# Check log file against known errors
# Return True if known error, False otherwise
def isknownerror(logfile):

    if not os.path.isfile(logfile):
        return False

    with open(logfile) as fp:
        for line in fp:
            line = line.strip()
            if re_known_errors.search(line):
                # Found a known issue
                return True

    return False


# Return how many times a commit hash / distro had combination has
# been retried for a given project
def timesretried(project, session, commit_hash, distro_hash):
    return session.query(Commit).filter(Commit.project_name == project,
                                        Commit.commit_hash == commit_hash,
                                        Commit.distro_hash == distro_hash,
                                        Commit.status == "RETRY").\
        count()


# Context manager to ensure locking for a file
@contextmanager
def lock_file(filename, mode='a'):
    with open(filename, mode) as lock_fp:
        fcntl.flock(lock_fp, fcntl.LOCK_EX)
        yield lock_fp
        fcntl.flock(lock_fp, fcntl.LOCK_UN)


# Run external pre-processing step
def run_external_preprocess(**kwargs):
    # Initially, get any params to be set as environment variables
    pkgname = kwargs.get('pkgname')
    distgit = kwargs.get('distgit')
    upstream_distgit = kwargs.get('upstream_distgit')
    cmdline = kwargs.get('cmdline')
    distroinfo = kwargs.get('distroinfo')
    srcdir = kwargs.get('source_dir')
    commit_hash = kwargs.get('commit_hash')
    username = os.environ.get('USER', None)

    run_cmd = []
    # Append environment variables
    if pkgname:
        run_cmd.append("DLRN_PACKAGE_NAME=%s" % pkgname)
    if distgit:
        run_cmd.append("DLRN_DISTGIT=%s" % distgit)
    if upstream_distgit:
        run_cmd.append("DLRN_UPSTREAM_DISTGIT=%s" % upstream_distgit)
    if distroinfo:
        run_cmd.append("DLRN_DISTROINFO_REPO=%s" % distroinfo)
    if srcdir:
        run_cmd.append("DLRN_SOURCEDIR=%s" % srcdir)
    if commit_hash:
        run_cmd.append("DLRN_SOURCE_COMMIT=%s" % commit_hash)
    if username:
        run_cmd.append("DLRN_USER=%s" % username)
    run_cmd.extend([cmdline])

    logger.info('Running custom pre-process: %s' % ' '.join(run_cmd))
    try:
        # We are forcing LANG to be C here, because env decides to use
        # non-ascii characters when the command is not found in UTF-8
        # environments
        sh.env(run_cmd, _cwd=distgit, _env={'LANG': 'C'})
    except Exception as e:
        msg = getattr(e, 'stderr', None)
        if msg:
            msg = msg.decode('utf-8')
        else:
            msg = e
        raise RuntimeError('Custom pre-process failed: %s' % msg)


# Return the list of all components that had a package built
def get_component_list(session):
    # The only way we have to get the components is to query the database
    all_comp_commits = session.query(Commit).\
        distinct(Commit.component).group_by(Commit.component).all()
    component_list = []
    for cmt in all_comp_commits:
        if cmt.component is not None:
            component_list.append(cmt.component)
    return component_list


# Aggregate all .repo files from a given symlink into a top-level repo file
# Also, aggregate the versions.csv file, this is useful for additional tooling
def aggregate_repo_files(dirname, datadir, session, reponame,
                         hashed_dir=False):
    component_list = get_component_list(session)

    repo_content = ''
    csv_content = []

    for component in component_list:
        repo_file = os.path.join(datadir, "repos/component", component,
                                 dirname, "%s.repo" % reponame)
        csv_file = os.path.join(datadir, "repos/component", component,
                                dirname, "versions.csv")

        if os.path.exists(repo_file):
            repo_content += open(repo_file).read() + '\n'
        if os.path.exists(csv_file):
            csv_content.extend(open(csv_file).readlines()[1:])

    file_hash = hashlib.md5(repo_content.encode()).hexdigest()

    if hashed_dir:
        target_dir = os.path.join(datadir, "repos", dirname, file_hash[:2],
                                  file_hash[2:4], file_hash)
    else:
        target_dir = os.path.join(datadir, "repos", dirname)

    # Create target directory if not present
    if not os.path.exists(target_dir):
        os.makedirs(target_dir)
    with open(os.path.join(target_dir, "%s.repo" % reponame), 'w') as fp:
        fp.write(repo_content)
    with open(os.path.join(target_dir, "%s.repo.md5" % reponame), 'w') as fp:
        fp.write(file_hash)
    with open(os.path.join(target_dir, "versions.csv"), 'w') as fp:
        fp.write("Project,Source Repo,Source Sha,Dist Repo,Dist Sha,"
                 "Status,Last Success Timestamp,Component,Extended Sha,"
                 "Pkg NVR\n")
        fp.writelines(csv_content)

    # If we created the file in a hashed dir, create the symlinks now
    if hashed_dir:
        base_promote_dir = os.path.join(datadir, "repos", dirname)
        os.symlink(os.path.relpath(os.path.join(target_dir,
                                                "%s.repo" % reponame),
                                   base_promote_dir),
                   os.path.join(base_promote_dir, "%s.repo_" % reponame))
        os.rename(os.path.join(base_promote_dir, "%s.repo_" % reponame),
                  os.path.join(base_promote_dir, "%s.repo" % reponame))
        os.symlink(os.path.relpath(os.path.join(target_dir,
                                                "%s.repo.md5" % reponame),
                                   base_promote_dir),
                   os.path.join(base_promote_dir, "%s.repo.md5_" % reponame))
        os.rename(os.path.join(base_promote_dir, "%s.repo.md5_" % reponame),
                  os.path.join(base_promote_dir, "%s.repo.md5" % reponame))
        os.symlink(os.path.relpath(os.path.join(target_dir, 'versions.csv'),
                                   base_promote_dir),
                   os.path.join(base_promote_dir, "versions.csv_"))
        os.rename(os.path.join(base_promote_dir, "versions.csv_"),
                  os.path.join(base_promote_dir, "versions.csv"))

    return file_hash


def find_in_artifacts(artifacts, word):
    if not artifacts:
        return
    for artifact in artifacts.split(','):
        if re.findall(word, artifact):
            return artifact


if __name__ == '__main__':
    s = getSession('sqlite:///%s' % sys.argv[1])
    saveYAML(s, sys.argv[1] + ".yaml")
    s = getSession('sqlite://')
    loadYAML(s, sys.argv[1] + ".yaml")
    print(s.query(Commit).first().project_name)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# BuildRPMDriver derived classes expose the following methods:
#
# build_package(). This method will perform the actual package build using
#                  the driver-specific approach.

from dlrn.drivers.buildrpm import BuildRPMDriver
import dlrn.shell

import io
import logging
import os
import re
import sh

logger = logging.getLogger("dlrn-build-mock")


class MockBuildDriver(BuildRPMDriver):
    DRIVER_CONFIG = {
        'mockbuild_driver': {
            'install_after_build': {'type': 'boolean', 'default': True},
        },
    }

    # We are using this method to "tee" mock output to mock.log and stdout
    def _process_mock_output(self, line):
        if dlrn.shell.verbose_build:
            logger.info(line[:-1])
        self.mock_fp.write(line)

    def __init__(self, *args, **kwargs):
        super(MockBuildDriver, self).__init__(*args, **kwargs)

    def build_package(self, **kwargs):
        """Valid parameters:

        :param output_directory: directory where the SRPM is located,
                                 and the built packages will be.
        :param additional_mock_opts: string with additional options to
                                     be passed to mock.
        """
        output_dir = kwargs.get('output_directory')
        additional_mock_opts = kwargs.get('additional_mock_opts')
        datadir = os.path.realpath(self.config_options.datadir)
        mock_config = os.environ.get('MOCK_CONFIG')
        install_after_build = self.config_options.install_after_build

        # Find src.rpm
        for rpm in os.listdir(output_dir):
            if rpm.endswith(".src.rpm"):
                src_rpm = '%s/%s' % (output_dir, rpm)
        try:
            # And build package
            with io.open("%s/mock.log" % output_dir, 'a',
                         encoding='utf-8', errors='replace') as self.mock_fp:
                try:
                    mock_opts = ['-v', '-r', '%s/%s' % (datadir, mock_config),
                                 '--resultdir', output_dir]
                    if additional_mock_opts:
                        mock_opts += [additional_mock_opts]
                    mock_opts += ['--rebuild', src_rpm]
                    sh.env('/usr/bin/mock', *mock_opts,
                           postinstall=install_after_build,
                           _err=self._process_mock_output,
                           _out=self._process_mock_output)
                except Exception as e:
                    raise e

            if install_after_build:
                # Check for warning about built packages failing to install
                with open("%s/mock.log" % output_dir, 'r') as fp:
                    mock_content = fp.readlines()
                warn_match = re.compile(
                    r'\W*WARNING: Failed install built packages.*')
                for line in mock_content:
                    m = warn_match.match(line)
                    if m is not None:
                        raise Exception('Failed to install built packages')

            # All went fine, create the $OUTPUT_DIRECTORY/installed file
            open('%s/installed' % output_dir, 'a').close()

        finally:
            with open("%s/mock.log" % output_dir, 'r') as fp:
                mock_content = fp.readlines()

            # Append mock output to rpmbuild.log
            with open('%s/rpmbuild.log' % output_dir, 'a') as fp:
                for line in mock_content:
                    fp.write(line)

            # Finally run restorecon
            try:
                sh.restorecon('-Rv', output_dir)
            except Exception as e:
                logger.info('restorecon did not run correctly, %s' % e)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os
import re
import sh
import shutil

from dlrn.config import getConfigOptions

logger = logging.getLogger("dlrn-repositories")


def refreshrepo(url, path, branch="master", local=False, full_path=None):
    config_options = getConfigOptions()
    logger.info("Getting %s to %s (%s)" % (url, path, branch))
    checkout_not_present = not os.path.exists(path)
    if checkout_not_present is True:
        try:
            sh.git.clone(url, path)
        except Exception as e:
            logger.error("Error cloning %s into %s: %s" % (url, path, e))
            raise

    elif local is False:
        # We need to cover a corner case here, where the repo URL has changed
        # since the last execution
        git = sh.git.bake(_cwd=path, _tty_out=False, _timeout=3600)
        try:
            remotes = git("remote", "-v").splitlines()
            fetch_url = None
            for remote in remotes:
                if '(fetch)' in remote:
                    line = remote.split()
                    if line[1] == url:
                        break
                    else:
                        fetch_url = line[1]
            else:
                # URL changed, so remove directory
                logger.warning("URL for %s changed from %s to %s, "
                               "cleaning directory and cloning again"
                               % (path, fetch_url, url))
                shutil.rmtree(path, ignore_errors=True)
                try:
                    sh.git.clone(url, path)
                except Exception as e:
                    logger.error("Error cloning %s into %s: %s" % (url, path,
                                                                   e))
                    raise
        except Exception:
            # Something failed here, maybe this is a failed repo clone
            # Let's warn, remove directory and clone again
            logger.warning("Directory %s does not contain a valid Git repo, "
                           "cleaning directory and cloning again" % path)
            shutil.rmtree(path)
            sh.git.clone(url, path)

    git_path = full_path or path
    git = sh.git.bake(_cwd=git_path, _tty_out=False, _timeout=3600)

    if local is False or checkout_not_present is True:
        try:
            git.fetch("origin")
        except Exception:
            # Sometimes hg repositories get into a invalid state leaving them
            # unusable, to avoid a looping error just remove it so it will be
            # recloned.
            logger.error("Error fetching into %s, deleting." % (path))
            sh.sudo("rm", "-rf", path)
            raise
        try:
            git.checkout('-f', branch)
        except sh.ErrorReturnCode_1:
            for branch_re in config_options.nonfallback_branches:
                if re.match(branch_re, branch):
                    # Do not try fallback on selected branches
                    raise
            else:
                if config_options.fallback_to_master:
                    # Fallback to master
                    if branch.startswith("rpm-"):
                        branch = "rpm-master"
                    elif branch.endswith("-rdo"):
                        # Distgit branches can start with rpm- or end with -rdo
                        branch = "rpm-master"
                    else:
                        branch = "master"
                    logger.info("Falling back to %s" % branch)
                    git.checkout(branch)
                else:
                    logger.error("Branch %s for %s does not exist, and the "
                                 "configuration does not allow a fallback to "
                                 "master." % (branch, url))
                    raise
        try:
            git.reset("--hard", "origin/%s" % branch)
        except Exception:
            # Maybe it was a tag, not a branch
            git.reset("--hard", "%s" % branch)

    repoinfo = str(git.log("--pretty=format:%H %ct", "-1", ".")).\
        strip().split(" ")
    repoinfo.insert(0, branch)
    return repoinfo


def getdistrobranch(package):
    if 'distro-branch' in package:
        return package['distro-branch']
    else:
        config_options = getConfigOptions()
        return config_options.distro


def getsourcebranch(package):
    if 'source-branch' in package:
        return package['source-branch']
    else:
        config_options = getConfigOptions()
        return config_options.source



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from __future__ import print_function
import argparse
import logging
import multiprocessing
import os
import sys
import tempfile

from copy import deepcopy
from functools import cmp_to_key
from functools import partial

import sh
from six.moves import configparser

from dlrn.build import build_worker

from dlrn.config import ConfigOptions
from dlrn.config import getConfigOptions
from dlrn.config import setup_logging

from dlrn.db import CIVote
from dlrn.db import closeSession
from dlrn.db import Commit
from dlrn.db import getCommits
from dlrn.db import getLastBuiltCommit
from dlrn.db import getLastProcessedCommit
from dlrn.db import getSession
from dlrn.db import Project
from dlrn.notifications import sendnotifymail
from dlrn.notifications import submit_review
from dlrn.reporting import genreports
from dlrn.repositories import getsourcebranch
from dlrn.rpmspecfile import RpmSpecCollection
from dlrn.rpmspecfile import RpmSpecFile
from dlrn.rsync import sync_repo
from dlrn.rsync import sync_symlinks
from dlrn.utils import aggregate_repo_files
from dlrn.utils import dumpshas2file
from dlrn.utils import import_object
from dlrn.utils import isknownerror
from dlrn.utils import lock_file
from dlrn.utils import saveYAML_commit
from dlrn.utils import timesretried
from dlrn import version

logger = logging.getLogger("dlrn")


def deprecation():
    # We will still call main, but will indicate that this way of calling
    # the application will be deprecated.
    print("Using the 'delorean' command has been deprecated. Please use 'dlrn'"
          " instead.")
    main()


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument('--config-file',
                        default='projects.ini',
                        help="Config file. Default: projects.ini")
    parser.add_argument('--config-override', action='append',
                        help="Override a configuration option from the"
                             " config file. Specify it as: "
                             "section.option=value. Can be used multiple "
                             "times if more than one override is needed.")
    parser.add_argument('--info-repo',
                        help="use a local distroinfo repo instead of"
                             " fetching the default one. Only applies when"
                             " pkginfo_driver is rdoinfo or downstream in"
                             " projects.ini")
    parser.add_argument('--build-env', action='append',
                        help="Variables for the build environment.")
    parser.add_argument('--local', action="store_true",
                        help="Use local git repos if possible. Only commited"
                             " changes in the local repo will be used in the"
                             " build.")
    parser.add_argument('--head-only', action="store_true",
                        help="Build from the most recent Git commit only.")
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--project-name', action='append',
                       help="Build a specific project name only."
                            " Use multiple times to build more than one "
                            "project in a run.")
    group.add_argument('--package-name', action='append',
                       help="Build a specific package name only."
                            " Use multiple times to build more than one "
                            "package in a run.")
    parser.add_argument('--dev', action="store_true",
                        help="Don't reset packaging git repo, force build "
                             "and add public master repo for dependencies "
                             "(dev mode).")
    parser.add_argument('--log-commands', action="store_true",
                        help="Log the commands run by dlrn.")
    parser.add_argument('--use-public', action="store_true",
                        help="Use the public master repo for dependencies "
                             "when doing install verification.")
    parser.add_argument('--order', action="store_true",
                        help="Compute the build order according to the spec "
                             "files instead of the dates of the commits. "
                             "Implies --sequential.")
    parser.add_argument('--sequential', action="store_true",
                        help="Run all actions sequentially, regardless of the"
                             " number of workers specified in projects.ini.")
    parser.add_argument('--status', action="store_true",
                        help="Get the status of packages.")
    parser.add_argument('--recheck', action="store_true",
                        help="Force a rebuild for a particular package. "
                        "Implies --package-name")
    parser.add_argument('--force-recheck', action="store_true",
                        help="Force a rebuild for a particular package, even "
                        "if its last build was successful. Requires setting "
                        "allow_force_rechecks=True in projects.ini. "
                        "Implies --package-name and --recheck")
    parser.add_argument('--version',
                        action='version',
                        version=version.version_info.version_string())
    parser.add_argument('--run',
                        help="Run a program instead of trying to build. "
                             "Implies --head-only")
    parser.add_argument('--stop', action="store_true",
                        help="Stop on error.")
    parser.add_argument('--verbose-build', action="store_true",
                        help="Show verbose output during the package build.")
    parser.add_argument('--verbose-mock', action="store_true",
                        help=argparse.SUPPRESS)
    parser.add_argument('--no-repo', action="store_true",
                        help="Do not generate a repo with all the built "
                        "packages.")
    parser.add_argument('--debug', action='store_true',
                        help="Print debug logs")

    options = parser.parse_args(sys.argv[1:])

    setup_logging(options.debug)

    if options.verbose_mock:
        logger.warning('The --verbose-mock command-line option is deprecated.'
                       ' Please use --verbose-build instead.')
        options.verbose_build = options.verbose_mock
    global verbose_build
    verbose_build = options.verbose_build

    cp = configparser.RawConfigParser()
    cp.read(options.config_file)

    if options.log_commands is True:
        logging.getLogger("sh.command").setLevel(logging.INFO)
    if options.order is True:
        options.sequential = True

    config_options = ConfigOptions(cp, overrides=options.config_override)
    if options.dev:
        _, tmpdb_path = tempfile.mkstemp()
        logger.info("Using file %s for temporary db" % tmpdb_path)
        config_options.database_connection = "sqlite:///%s" % tmpdb_path

    session = getSession(config_options.database_connection)
    pkginfo_driver = config_options.pkginfo_driver
    global pkginfo
    pkginfo = import_object(pkginfo_driver, cfg_options=config_options)
    packages = pkginfo.getpackages(local_info_repo=options.info_repo,
                                   tags=config_options.tags,
                                   dev_mode=options.dev)

    if options.project_name:
        pkg_names = [p['name'] for p in packages
                     if p['project'] in options.project_name]
    elif options.package_name:
        pkg_names = options.package_name
    else:
        pkg_names = None

    if options.status is True:
        if not pkg_names:
            pkg_names = [p['name'] for p in packages]
        for name in pkg_names:
            package = [p for p in packages if p['name'] == name][0]
            for build_type in package.get('types', ['rpm']):
                commit = getLastProcessedCommit(
                    session, name, 'invalid status',
                    type=build_type)
                if commit:
                    print("{:>9}".format(build_type), name, commit.status)
                else:
                    print("{:>9}".format(build_type), name, 'NO_BUILD')
        sys.exit(0)

    if pkg_names:
        pkg_name = pkg_names[0]
    else:
        pkg_name = None

    def recheck_commit(commit, force):
        if commit.status == 'SUCCESS':
            if not force:
                logger.error(
                    "Trying to recheck an already successful commit,"
                    " ignoring. If you want to force it, use --force-recheck"
                    " and set allow_force_rechecks=True in projects.ini")
                sys.exit(1)
            else:
                logger.info("Forcefully rechecking a successfully built "
                            "commit for %s" % commit.project_name)
        elif commit.status == 'RETRY':
            # In this case, we are going to retry anyway, so
            # do nothing and exit
            logger.warning("Trying to recheck a commit in RETRY state,"
                           " ignoring.")
            sys.exit(0)
        # We could set the status to RETRY here, but if we have gone
        # beyond max_retries it wouldn't work as expected. Thus, our
        # only chance is to remove the commit
        session.delete(commit)
        session.commit()
        sys.exit(0)

    if options.recheck is True:
        if not pkg_name:
            logger.error('Please use --package-name or --project-name '
                         'with --recheck.')
            sys.exit(1)

        if options.force_recheck and config_options.allow_force_rechecks:
            force_recheck = True
        else:
            force_recheck = False
        package = [p for p in packages if p['name'] == pkg_name][0]
        for build_type in package.get('types', ['rpm']):
            commit = getLastProcessedCommit(session, pkg_name, type=build_type)
            if commit:
                recheck_commit(commit, force_recheck)
            else:
                logger.error("There are no existing commits for package %s",
                             pkg_name)
                sys.exit(1)
    # when we run a program instead of building we don't care about
    # the commits, we just want to run once per package
    if options.run:
        options.head_only = True
    # Build a list of commits we need to process
    toprocess = []

    def add_commits(project_toprocess):
        # The first entry in the list of commits is a commit we have
        # already processed, we want to process it again only if in dev
        # mode or distro hash has changed, we can't simply check
        # against the last commit in the db, as multiple commits can
        # have the same commit date
        for commit_toprocess in project_toprocess:
            if options.dev is True or \
               options.run or \
               not session.query(Commit).filter(
                   Commit.commit_hash == commit_toprocess.commit_hash,
                   Commit.distro_hash == commit_toprocess.distro_hash,
                   Commit.extended_hash == commit_toprocess.extended_hash,
                   Commit.type == commit_toprocess.type,
                   Commit.status != "RETRY").all():
                toprocess.append(commit_toprocess)

    if not pkg_name and not pkg_names:
        pool = multiprocessing.Pool()   # This will use all the system cpus
        # Use functools.partial to iterate on the packages to process,
        # while keeping a few options fixed
        getinfo_wrapper = partial(getinfo, local=options.local,
                                  dev_mode=options.dev,
                                  head_only=options.head_only,
                                  db_connection=config_options.
                                  database_connection)
        iterator = pool.imap(getinfo_wrapper, packages)
        while True:
            try:
                project_toprocess, updated_pkg = iterator.next()
                for package in packages:
                    if package['name'] == updated_pkg['name']:
                        if package['upstream'] == 'Unknown':
                            package['upstream'] = updated_pkg['upstream']
                            logger.debug(
                                "Updated upstream for package %s to %s",
                                package['name'], package['upstream'])
                        break
                add_commits(project_toprocess)
            except StopIteration:
                break
        pool.close()
        pool.join()
    else:
        for package in packages:
            if package['name'] in pkg_names:
                project_toprocess, _ = getinfo(package, local=options.local,
                                               dev_mode=options.dev,
                                               head_only=options.head_only,
                                               db_connection=config_options.
                                               database_connection)
                add_commits(project_toprocess)
    closeSession(session)   # Close session, will reopen during post_build

    # Check if there is any commit at all to process
    if len(toprocess) == 0:
        if not pkg_name:
            # Use a shorter message if this was a full run
            logger.info("No commits to build.")
        else:
            logger.info("No commits to build. If this is not expected, please"
                        " make sure the package name(s) are correct, and that "
                        "any failed commit you want to rebuild has been "
                        "removed from the database.")
        return 0

    # if requested do a sort according to build and install
    # dependencies
    if options.order is True:
        # collect info from all spec files
        logger.info("Reading rpm spec files")
        projects = sorted([c.project_name for c in toprocess])

        speclist = []
        bootstraplist = []
        for project_name in projects:
            # Preprocess spec if needed
            pkginfo.preprocess(package_name=project_name)

            filename = None
            for f in os.listdir(pkginfo.distgit_dir(project_name)):
                if f.endswith('.spec'):
                    filename = f

            if filename:
                specpath = os.path.join(pkginfo.distgit_dir(project_name),
                                        filename)
                speclist.append(sh.rpmspec('-D', 'repo_bootstrap 1',
                                           '-P', specpath))
                # Check if repo_bootstrap is defined in the package.
                # If so, we'll need to rebuild after the whole bootstrap
                rawspec = open(specpath).read(-1)
                if 'repo_bootstrap' in rawspec:
                    bootstraplist.append(project_name)
            else:
                logger.warning("Could not find a spec for package %s" %
                               project_name)

        logger.debug("Packages to rebuild: %s" % bootstraplist)

        specs = RpmSpecCollection([RpmSpecFile(spec)
                                  for spec in speclist])
        # compute order according to BuildRequires
        logger.info("Computing build order")
        orders = specs.compute_order()
        # hack because the package name is not consistent with the directory
        # name and the spec file name
        if 'python-networking_arista' in orders:
            orders.insert(orders.index('python-networking_arista'),
                          'python-networking-arista')

        # sort the commits according to the score of their project and
        # then use the timestamp of the commits as a secondary key
        def my_cmp(a, b):
            if a.project_name == b.project_name:
                _a = a.dt_commit
                _b = b.dt_commit
            else:
                _a = orders.index(a.project_name) if a.project_name in \
                    orders else sys.maxsize
                _b = orders.index(b.project_name) if b.project_name in \
                    orders else sys.maxsize
            # cmp is no longer available in python3 so replace it. See Ordering
            # Comparisons on:
            # https://docs.python.org/3.0/whatsnew/3.0.html
            return (_a > _b) - (_a < _b)

        toprocess.sort(key=cmp_to_key(my_cmp))
    else:
        # sort according to the timestamp of the commits
        toprocess.sort()

    exit_code = 0
    if options.sequential is True:
        toprocess_copy = deepcopy(toprocess)
        for commit in toprocess:
            status = build_worker(packages, commit, run_cmd=options.run,
                                  build_env=options.build_env,
                                  dev_mode=options.dev,
                                  use_public=options.use_public,
                                  order=options.order, sequential=True)
            exception = status[3]
            consistent = False
            datadir = os.path.realpath(config_options.datadir)
            with lock_file(os.path.join(datadir, 'remote.lck')):
                session = getSession(config_options.database_connection)
                if exception is not None:
                    logger.error("Received exception %s" % exception)
                    failures = 1
                else:
                    if not options.run:
                        failures = post_build(status, packages, session,
                                              build_repo=not options.no_repo)
                        consistent = (failures == 0)
                exit_value = process_build_result(status, packages, session,
                                                  toprocess_copy,
                                                  dev_mode=options.dev,
                                                  run_cmd=options.run,
                                                  stop=options.stop,
                                                  build_env=options.build_env,
                                                  head_only=options.head_only,
                                                  consistent=consistent,
                                                  failures=failures)
                closeSession(session)

            if exit_value != 0:
                exit_code = exit_value
            if options.stop and exit_code != 0:
                return exit_code
    else:
        # Setup multiprocessing pool
        pool = multiprocessing.Pool(config_options.workers)
        # Use functools.partial to iterate on the commits to process,
        # while keeping a few options fixed
        build_worker_wrapper = partial(build_worker, packages,
                                       run_cmd=options.run,
                                       build_env=options.build_env,
                                       dev_mode=options.dev,
                                       use_public=options.use_public,
                                       order=options.order, sequential=False)
        iterator = pool.imap(build_worker_wrapper, toprocess)

        while True:
            try:
                status = iterator.next()
                exception = status[3]
                consistent = False
                datadir = os.path.realpath(config_options.datadir)
                with lock_file(os.path.join(datadir, 'remote.lck')):
                    session = getSession(config_options.database_connection)
                    if exception is not None:
                        logger.info("Received exception %s" % exception)
                        failures = 1
                    else:
                        # Create repo, build versions.csv file.
                        # This needs to be sequential
                        if not options.run:
                            failures = post_build(
                                status, packages, session,
                                build_repo=not options.no_repo)
                            consistent = (failures == 0)
                    exit_value = process_build_result(
                        status, packages,
                        session, toprocess,
                        dev_mode=options.dev,
                        run_cmd=options.run,
                        stop=options.stop,
                        build_env=options.build_env,
                        head_only=options.head_only,
                        consistent=consistent,
                        failures=failures)
                    closeSession(session)
                if exit_value != 0:
                    exit_code = exit_value
                if options.stop and exit_code != 0:
                    return exit_code
            except StopIteration:
                break
        pool.close()
        pool.join()

    # If we were bootstrapping, set the packages that required it to RETRY
    session = getSession(config_options.database_connection)
    if options.order is True and not pkg_name:
        for bpackage in bootstraplist:
            commit = getLastProcessedCommit(session, bpackage)
            commit.status = 'RETRY'
            session.add(commit)
            session.commit()
    genreports(packages, options.head_only, session, [])
    closeSession(session)

    if options.dev:
        os.remove(tmpdb_path)
    return exit_code


def process_build_result(status, *args, **kwargs):
    if status[0].type == "rpm":
        return process_build_result_rpm(status, *args, **kwargs)
    elif status[0].type == "container":
        return process_build_result_container(status, *args, **kwargs)
    else:
        raise Exception("Unknown type %s" % status[0].type)


def process_build_result_container(
        status, packages, session, packages_to_process,
        dev_mode=False, run_cmd=False, stop=False,
        build_env=None, head_only=False, consistent=False,
        failures=0):
    raise NotImplementedError()


def process_build_result_rpm(
        status, packages, session, packages_to_process,
        dev_mode=False, run_cmd=False, stop=False,
        build_env=None, head_only=False, consistent=False,
        failures=0):
    config_options = getConfigOptions()
    commit = status[0]
    built_rpms = status[1]
    notes = status[2]
    exception = status[3]
    commit_hash = commit.commit_hash
    project = commit.project_name
    project_info = session.query(Project).filter(
        Project.project_name == project).first()
    if not project_info:
        project_info = Project(project_name=project, last_email=0)
    exit_code = 0

    if run_cmd:
        if exception is not None:
            exit_code = 1
            if stop:
                return exit_code
        return exit_code

    if exception is None:
        commit.status = "SUCCESS"
        commit.notes = notes
        commit.artifacts = ",".join(built_rpms)
    else:
        logger.error("Received exception %s" % exception)

        datadir = os.path.realpath(config_options.datadir)
        yumrepodir = os.path.join(datadir, "repos",
                                  commit.getshardedcommitdir())
        logfile = os.path.join(yumrepodir,
                               "rpmbuild.log")

        # If the log file hasn't been created we add what we have
        # This happens if the rpm build script didn't run.
        if not os.path.exists(yumrepodir):
            os.makedirs(yumrepodir)
        if not os.path.exists(logfile):
            with open(logfile, "w") as fp:
                fp.write(str(exception))

        if (isknownerror(logfile) and
            (timesretried(project, session, commit_hash, commit.distro_hash) <
             config_options.maxretries)):
            logger.exception("Known error building packages for %s,"
                             " will retry later" % project)
            commit.status = "RETRY"
            commit.notes = str(exception)
            # do not switch from an error exit code to a retry
            # exit code
            if exit_code != 1:
                exit_code = 2
        else:
            exit_code = 1

            if not project_info.suppress_email():
                sendnotifymail(packages, commit)
                project_info.sent_email()
                session.add(project_info)

            # allow to submit a gerrit review only if the last build
            # was successful or non existent to avoid creating a gerrit
            # review for the same problem multiple times.
            if config_options.gerrit is not None:
                if build_env:
                    env_vars = list(build_env)
                else:
                    env_vars = []
                last_build = getLastProcessedCommit(session, project)
                if not last_build or last_build.status == 'SUCCESS':
                    try:
                        submit_review(commit, packages, env_vars)
                    except Exception:
                        logger.error('Unable to create review '
                                     'see review.log')
                else:
                    logger.info('Last build not successful '
                                'for %s' % project)
            commit.status = "FAILED"
            commit.notes = str(exception)
        if stop:
            return exit_code
    # Add commit to the session
    session.add(commit)

    genreports(packages, head_only, session, packages_to_process)
    # Export YAML file containing commit metadata
    export_commit_yaml(commit)
    try:
        sync_repo(commit)
    except Exception as e:
        logger.error('Repo sync failed for project %s' % project)
        consistent = False  # If we were consistent before, we are not anymore
        if exit_code == 0:  # The commit was ok, so marking as failed
            exit_code = 1
            # We need to make the commit status be "failed"
            commit.status = "FAILED"
            commit.notes = str(e)
            session.add(commit)
            # And open a review if needed
            if config_options.gerrit is not None:
                if build_env:
                    env_vars = list(build_env)
                else:
                    env_vars = []
                try:
                    submit_review(commit, packages, env_vars)
                except Exception:
                    logger.error('Unable to create review '
                                 'see review.log')

    session.commit()

    # Generate the current and consistent symlinks
    if exception is None:
        dirnames = ['current']
        datadir = os.path.realpath(config_options.datadir)
        yumrepodir = os.path.join(datadir, "repos",
                                  commit.getshardedcommitdir())
        yumrepodir_abs = os.path.join(datadir, yumrepodir)
        if consistent:
            dirnames.append('consistent')
        else:
            if config_options.use_components:
                logger.info('%d packages not built correctly for component'
                            ' %s: not updating the consistent symlink' %
                            (failures, commit.component))
            else:
                logger.info('%d packages not built correctly: not updating'
                            ' the consistent symlink' % failures)
        for dirname in dirnames:
            if config_options.use_components:
                target_repo_dir = os.path.join(datadir, "repos/component",
                                               commit.component, dirname)
                source_repo_dir = os.path.join(datadir, "repos/component",
                                               commit.component)
            else:
                target_repo_dir = os.path.join(datadir, "repos", dirname)
                source_repo_dir = os.path.join(datadir, "repos")
            os.symlink(os.path.relpath(yumrepodir_abs, source_repo_dir),
                       target_repo_dir + "_")
            os.rename(target_repo_dir + "_", target_repo_dir)

        # If using components, synchronize the upper-level repo files
        if config_options.use_components:
            for dirname in dirnames:
                aggregate_repo_files(dirname, datadir, session,
                                     config_options.reponame, hashed_dir=True)

        # And synchronize them
        sync_symlinks(commit)

    if dev_mode is False:
        if consistent:
            # We have a consistent repo. Let's create a CIVote entry in the DB
            vote = CIVote(commit_id=commit.id, ci_name='consistent',
                          ci_url='', ci_vote=True, ci_in_progress=False,
                          timestamp=int(commit.dt_build), notes='',
                          component=commit.component)
            session.add(vote)
            session.commit()
    return exit_code


def export_commit_yaml(commit):
    config_options = getConfigOptions()
    # Export YAML file containing commit metadata
    datadir = os.path.realpath(config_options.datadir)
    yumrepodir = os.path.join(datadir, "repos",
                              commit.getshardedcommitdir())
    saveYAML_commit(commit, os.path.join(yumrepodir, 'commit.yaml'))


def post_build(status, *args, **kwargs):
    if status[0].type == "rpm":
        return post_build_rpm(status, *args, **kwargs)
    elif status[0].type == "container":
        return post_build_container(status, *args, **kwargs)
    else:
        raise Exception("Unknown type %s" % status[0].type)


def post_build_container(status, packages, session, build_repo=None):
    raise NotImplementedError()


def post_build_rpm(status, packages, session, build_repo=True):
    config_options = getConfigOptions()
    commit = status[0]
    built_rpms = status[1]
    project_name = commit.project_name
    commit_hash = commit.commit_hash
    datadir = os.path.realpath(config_options.datadir)
    yumrepodir = os.path.join("repos", commit.getshardedcommitdir())
    yumrepodir_abs = os.path.join(datadir, yumrepodir)

    shafile = open(os.path.join(yumrepodir_abs, "versions.csv"), "w")
    shafile.write("Project,Source Repo,Source Sha,Dist Repo,Dist Sha,"
                  "Status,Last Success Timestamp,Component,Extended Sha,"
                  "Pkg NVR\n")
    failures = 0

    for otherproject in packages:
        if (config_options.use_components and 'component' in otherproject and
                otherproject['component'] != commit.component):
            # Only dump information and create symlinks for the same component
            continue

        otherprojectname = otherproject["name"]
        if otherprojectname == project_name:
            # Output sha's this project
            dumpshas2file(shafile, commit, otherproject["upstream"],
                          otherproject["master-distgit"], "SUCCESS",
                          commit.dt_build, commit.component, built_rpms)
            continue
        # Output sha's of all other projects represented in this repo
        last_success = getCommits(session, project=otherprojectname,
                                  with_status="SUCCESS",
                                  type=commit.type).first()
        last_processed = getCommits(session, project=otherprojectname,
                                    type=commit.type).first()

        if last_success:
            if build_repo:
                for rpm in last_success.artifacts.split(","):
                    rpm_link_src = os.path.join(yumrepodir_abs,
                                                os.path.split(rpm)[1])
                    os.symlink(os.path.relpath(os.path.join(datadir, rpm),
                                               yumrepodir_abs), rpm_link_src)
            last = last_success
        else:
            last = last_processed
        if last:
            if last.artifacts:
                rpmlist = last.artifacts.split(",")
            else:
                rpmlist = []
            upstream = otherproject.get('upstream', '')
            dumpshas2file(shafile, last, upstream,
                          otherproject["master-distgit"],
                          last_processed.status, last.dt_build,
                          commit.component, rpmlist)
            if last_processed.status != 'SUCCESS':
                failures += 1
        else:
            failures += 1
    shafile.close()

    if build_repo:
        # Use createrepo_c when available
        try:
            from sh import createrepo_c
            sh.createrepo = createrepo_c
        except ImportError:
            pass

        if config_options.include_srpm_in_repo:
            sh.createrepo(yumrepodir_abs)
        else:
            sh.createrepo('-x', '*.src.rpm', yumrepodir_abs)

        with open(os.path.join(
                yumrepodir_abs, "%s.repo" % config_options.reponame),
                "w") as fp:
            if config_options.use_components:
                repo_id = "%s-component-%s" % (config_options.reponame,
                                               commit.component)
            else:
                repo_id = config_options.reponame
            fp.write("[%s]\nname=%s-%s-%s\nbaseurl=%s/%s\nenabled=1\n"
                     "gpgcheck=0\npriority=1\n" % (
                         repo_id,
                         config_options.reponame,
                         project_name, commit_hash,
                         config_options.baseurl,
                         commit.getshardedcommitdir()))

    return failures


def getinfo(package, local=False, dev_mode=False, head_only=False,
            db_connection=None, type="rpm"):
    project = package["name"]
    since = "-1"
    session = getSession(db_connection)
    commit = getLastProcessedCommit(session, project, type=type)
    if commit:
        # If we have switched source branches, we want to behave
        # as if no previous commits had been built, and only build
        # the last one
        if commit.commit_branch == getsourcebranch(package):
            # This will return all commits since the last handled commit
            # including the last handled commit, remove it later if needed.
            since = "--after=%d" % (commit.dt_commit)
        else:
            # The last processed commit belongs to a different branch. Just
            # in case, let's check if we built a previous commit from the
            # current branch
            commit = getLastBuiltCommit(session, project,
                                        getsourcebranch(package), type=type)
            if commit:
                logger.info("Last commit belongs to another branch, but"
                            " we're ok with that")
                since = "--after=%d" % (commit.dt_commit)
                # In any case, we just want to build the last commit, if any
                head_only = True

    project_toprocess = pkginfo.getinfo(project=project, package=package,
                                        since=since, local=local,
                                        dev_mode=dev_mode, type=type)

    closeSession(session)
    # If since == -1, then we only want to trigger a build for the
    # most recent change
    if since == "-1" or head_only:
        del project_toprocess[:-1]

    return project_toprocess, package



#
# Copyright (C) 2015 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import unittest

from dlrn.rpmspecfile import RpmSpecCollection
from dlrn.rpmspecfile import RpmSpecFile


class TestRpmSpecFile(unittest.TestCase):

    def test_basic_package(self):
        spec = RpmSpecFile(BASIC_SPEC_CONTENT)
        self.assertEqual(spec.packages(), ['package'])

    def test_package_with_macro(self):
        spec = RpmSpecFile(MACRO_SPEC_CONTENT)
        self.assertEqual(spec.packages(), ['package'])

    def test_package_with_provides(self):
        spec = RpmSpecFile(PROVIDES_SPEC_CONTENT)
        self.assertEqual(spec.packages(), ['package', 'oldname'])

    def test_sub_package(self):
        spec = RpmSpecFile(SUB_PKG_CONTENT)
        self.assertEqual(spec.packages(), ['package', 'package-subpkg-toto'])

    def test_nsub_package(self):
        spec = RpmSpecFile(NSUB_PKG_CONTENT)
        self.assertEqual(spec.packages(), ['package', 'pre-serv-post'])

    def test_name_package(self):
        spec = RpmSpecFile(NAME_PKG_CONTENT)
        self.assertEqual(spec.packages(), ['package', 'package-package-toto'])

    def test_build_requires(self):
        spec = RpmSpecFile(DEP_SPEC_CONTENT)
        self.assertEqual(set(spec.build_requires()), set(['dep1', 'dep2']))

    def test_build_requires_with_operator(self):
        spec = RpmSpecFile(OPERATOR_DEP_SPEC_CONTENT)
        self.assertEqual(set(spec.build_requires()), set(['dep1', 'dep2']))

    def test_build_requires_with_operator_no_space(self):
        spec = RpmSpecFile(OPERATOR_DEP_SPEC_CONTENT2)
        self.assertEqual(set(spec.build_requires()), set(['dep1', 'dep2']))


class TestRpmSpecCollection(unittest.TestCase):

    def test_basic(self):
        specs = RpmSpecCollection([RpmSpecFile(BASIC_SPEC_CONTENT)])
        self.assertEqual(specs.compute_order(), ['package'])

    def test_multiple_names(self):
        specs = RpmSpecCollection([RpmSpecFile(MULTIPLE_NAME_CONTENT)])
        self.assertEqual(specs.compute_order(), ['package'])

    def test_dep(self):
        specs = RpmSpecCollection([RpmSpecFile(BASIC_SPEC_CONTENT),
                                   RpmSpecFile(BASIC2_SPEC_CONTENT)])
        self.assertEqual(specs.compute_order(), ['package', 'packageC'])

    def test_dep2(self):
        specs = RpmSpecCollection([RpmSpecFile(BASIC_SPEC_CONTENT),
                                  RpmSpecFile(BASIC2_SPEC_CONTENT),
                                   RpmSpecFile(BASIC3_SPEC_CONTENT)])
        self.assertEqual(specs.compute_order(), ['package',
                                                 'packageC',
                                                 'packageB'])

    def test_dep_sub(self):
        specs = RpmSpecCollection([RpmSpecFile(DEP_SUB_PKG_CONTENT),
                                  RpmSpecFile(SUB_PKG_CONTENT)])
        self.assertEqual(specs.compute_order(), ['package',
                                                 'packageD'])


BASIC_SPEC_CONTENT = '''
Name: package
'''

BASIC2_SPEC_CONTENT = '''
Name: packageC
Requires(pre): package
'''

BASIC3_SPEC_CONTENT = '''
Name: packageB
BuildRequires: packageC, unknown
'''

MACRO_SPEC_CONTENT = '''
%define name package

Name: %{name}
'''

PROVIDES_SPEC_CONTENT = '''
Name: package
Provides: oldname = 1.0
'''

DEP_SPEC_CONTENT = '''
Name: package
BuildRequires: dep1, dep2
'''

OPERATOR_DEP_SPEC_CONTENT = '''
Name: package
BuildRequires: dep1, dep2 >= 0.1
'''

OPERATOR_DEP_SPEC_CONTENT2 = '''
Name: package
BuildRequires: dep1, dep2>=0.1
'''

SUB_PKG_CONTENT = '''
Name: package
%define serv subpkg
%package %{serv}-toto
'''

NSUB_PKG_CONTENT = '''
Name: package
%global sub serv
%global serv %{sub}
%package -n pre-%{serv}-post
'''

SUB_PKG_CONTENT = '''
Name: package
%define serv subpkg
%package %{serv}-toto
'''

NAME_PKG_CONTENT = '''
Name: package
%package %{name}-toto
'''

DEP_SUB_PKG_CONTENT = '''
Name: packageD
BuildRequires: package-subpkg-toto
'''

MULTIPLE_NAME_CONTENT = '''
Name: package

%prep
cat > fake.egg-info <<EOF
Name: anothername
EOF
'''

if __name__ == "__main__":
    unittest.main()

# test_rpmspecfile.py ends here



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Extra tables for votes on aggregates

Revision ID: 7fbd3a18502f
Revises: 00a31f1f39c0
Create Date: 2020-01-16 15:22:32.090726

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '7fbd3a18502f'
down_revision = '00a31f1f39c0'
branch_labels = None
depends_on = None


def upgrade():
    # Since SQLite3 does not allow ALTER TABLE statements, we need to do a
    # batch operation: http://alembic.zzzcomputing.com/en/latest/batch.html
    with op.batch_alter_table('promotions') as batch_op:
        batch_op.add_column(sa.Column('aggregate_hash', sa.String(64),
                                      nullable=True))

    op.create_table('civotes_agg',
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('ref_hash', sa.String(64), nullable=False),
                    sa.Column('ci_name', sa.String(256), nullable=True),
                    sa.Column('ci_url', sa.String(1024), nullable=True),
                    sa.Column('ci_vote', sa.Boolean(), nullable=True),
                    sa.Column('ci_in_progress', sa.Boolean(), nullable=True),
                    sa.Column('timestamp', sa.Integer(), nullable=True),
                    sa.Column('notes', sa.Text(), nullable=True),
                    sa.Column('user', sa.String(255),
                              sa.ForeignKey('users.username')),
                    sa.PrimaryKeyConstraint('id'))


def downgrade():
    op.drop_table('civotes_agg')
    with op.batch_alter_table('promotions') as batch_op:
        batch_op.drop_column('aggregate_hash')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Initial creation

Revision ID: 3c62b0d3ec34
Revises:
Create Date: 2016-06-17 10:35:54.703517

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '3c62b0d3ec34'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    try:
        op.create_table('commits',
                        sa.Column('id', sa.Integer(), nullable=False),
                        sa.Column('dt_commit', sa.Integer(), nullable=True),
                        sa.Column('dt_distro', sa.Integer(), nullable=True),
                        sa.Column('dt_build', sa.Integer(), nullable=True),
                        sa.Column('project_name', sa.String(), nullable=True),
                        sa.Column('repo_dir', sa.String(), nullable=True),
                        sa.Column('commit_hash', sa.String(), nullable=True),
                        sa.Column('distro_hash', sa.String(), nullable=True),
                        sa.Column('status', sa.String(), nullable=True),
                        sa.Column('rpms', sa.String(), nullable=True),
                        sa.Column('notes', sa.String(), nullable=True),
                        sa.Column('flags', sa.Integer(), nullable=True),
                        sa.PrimaryKeyConstraint('id'))
        op.create_table('projects',
                        sa.Column('id', sa.Integer(), nullable=False),
                        sa.Column('project_name', sa.String(), nullable=True),
                        sa.Column('last_email', sa.Integer(), nullable=True),
                        sa.PrimaryKeyConstraint('id'))
    except Exception:
        # this means that's a legacy table unversioned by alembic
        pass


def downgrade():
    op.drop_table('projects')
    op.drop_table('commits')



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
import os
import sh
import shutil
import tempfile

from six.moves import configparser

import dlrn.shell

from dlrn.build import build
from dlrn.build import build_rpm_wrapper
from dlrn.config import ConfigOptions
from dlrn import db
from dlrn.tests import base
from dlrn import utils


class FakePkgInfo(object):
    def preprocess(self, **argv):
        return


dlrn.shell.pkginfo = FakePkgInfo()


def mocked_listdir(directory):
    return ['python-pysaml2-3.0-1a.el7.centos.src.rpm']


@mock.patch('sh.restorecon', create=True)
@mock.patch('sh.env', create=True)
@mock.patch.object(sh.Command, '__call__', autospec=True)
class TestBuild(base.TestCase):
    def setUp(self):
        super(TestBuild, self).setUp()
        self.configfile = configparser.RawConfigParser()
        self.configfile.read("projects.ini")
        self.configfile.set('DEFAULT', 'datadir', tempfile.mkdtemp())
        self.configfile.set('DEFAULT', 'scriptsdir', tempfile.mkdtemp())
        self.configfile.set('DEFAULT', 'baseurl', "file://%s" %
                            self.configfile.get('DEFAULT', 'datadir'))
        self.config = ConfigOptions(self.configfile)
        shutil.copyfile(os.path.join("scripts", "centos8.cfg"),
                        os.path.join(self.config.scriptsdir, "centos8.cfg"))
        with open(os.path.join(self.config.datadir,
                  "delorean-deps.repo"), "w") as fp:
            fp.write("[test]\nname=test\nenabled=0\n")
        self.db_fd, filepath = tempfile.mkstemp()
        self.session = db.getSession("sqlite:///%s" % filepath)
        utils.loadYAML(self.session, './dlrn/tests/samples/commits_1.yaml')

    def tearDown(self):
        super(TestBuild, self).tearDown()
        shutil.rmtree(self.config.datadir)
        shutil.rmtree(self.config.scriptsdir)
        os.close(self.db_fd)
        # Make sure env vars are cleaned up
        if 'RELEASE_NUMBERING' in os.environ:
            del os.environ['RELEASE_NUMBERING']
        if 'RELEASE_MINOR' in os.environ:
            del os.environ['RELEASE_MINOR']

    @mock.patch('os.listdir', side_effect=mocked_listdir)
    def test_build_rpm_wrapper(self, ld_mock, sh_mock, env_mock, rc_mock):
        self.configfile.set('DEFAULT', 'build_driver',
                            'dlrn.drivers.mockdriver.MockBuildDriver')
        self.config = ConfigOptions(self.configfile)
        commit = db.getCommits(self.session)[-1]
        build_rpm_wrapper(commit, False, False, False, None, True)
        # 3 sh calls:
        # 1- build_srpm.sh
        # 2- mock (handled by env_mock)
        # 3- restorecon (handled by rc_mock)
        self.assertEqual(env_mock.call_count, 2)
        self.assertEqual(rc_mock.call_count, 1)
        self.assertTrue(os.path.exists(os.path.join(self.config.datadir,
                                                    "dlrn-1.cfg")))

    @mock.patch('os.listdir', side_effect=mocked_listdir)
    def test_build_rpm_wrapper_release_numbering(self, ld_mock, sh_mock,
                                                 env_mock, rc_mock):
        self.configfile.set('DEFAULT', 'build_driver',
                            'dlrn.drivers.mockdriver.MockBuildDriver')
        self.configfile.set('DEFAULT', 'release_numbering',
                            'minor.date.hash')
        self.configfile.set('DEFAULT', 'release_minor',
                            '2')
        self.config = ConfigOptions(self.configfile)
        commit = db.getCommits(self.session)[-1]
        build_rpm_wrapper(commit, False, False, False, None, True)
        self.assertEqual(os.environ['RELEASE_NUMBERING'], 'minor.date.hash')
        self.assertEqual(os.environ['RELEASE_MINOR'], '2')

    @mock.patch('os.listdir', side_effect=mocked_listdir)
    def test_build(self, ld_mock, sh_mock, env_mock, rc_mock):
        self.configfile.set('DEFAULT', 'build_driver',
                            'dlrn.drivers.mockdriver.MockBuildDriver')
        self.config = ConfigOptions(self.configfile)
        commit = db.getCommits(self.session)[-1]
        try:
            build([], commit, None, False, False, False, True)
        except Exception as e:
            self.assertIn("No rpms built for", str(e))

    @mock.patch('os.listdir', side_effect=mocked_listdir)
    def test_build_configdir(self, ld_mock, sh_mock, env_mock, rc_mock):
        configdir = tempfile.mkdtemp()
        self.configfile.set('DEFAULT', 'configdir', configdir)
        self.configfile.set('DEFAULT', 'build_driver',
                            'dlrn.drivers.mockdriver.MockBuildDriver')
        self.config = ConfigOptions(self.configfile)

        shutil.copyfile(os.path.join("scripts", "centos8.cfg"),
                        os.path.join(configdir, "centos8.cfg"))
        commit = db.getCommits(self.session)[-1]
        expected = [mock.call('%s/centos8.cfg' % configdir,
                              '%s/dlrn-1.cfg.new' % self.config.datadir),
                    mock.call('%s/dlrn-1.cfg.new' % self.config.datadir,
                              '%s/dlrn-1.cfg' % self.config.datadir)]

        with mock.patch('shutil.copyfile',
                        side_effect=shutil.copyfile) as cp_mock:
            build_rpm_wrapper(commit, False, False, False, None, True)
            self.assertEqual(expected, cp_mock.call_args_list)

    @mock.patch('dlrn.drivers.kojidriver.KojiBuildDriver.build_package')
    @mock.patch('os.listdir', side_effect=mocked_listdir)
    @mock.patch('dlrn.drivers.kojidriver.KojiBuildDriver.write_mock_config',
                create=True)
    def test_build_rpm_wrapper_mock_config(self, wm_mock, ld_mock, bp_mock,
                                           sh_mock, env_mock, rc_mock):
        self.configfile.set('kojibuild_driver', 'fetch_mock_config', 'True')
        self.configfile.set('DEFAULT', 'build_driver',
                            'dlrn.drivers.kojidriver.KojiBuildDriver')
        self.config = ConfigOptions(self.configfile)
        commit = db.getCommits(self.session)[-1]
        build_rpm_wrapper(commit, False, False, False, None, True)
        self.assertEqual(wm_mock.call_count, 1)



#!/usr/bin/env python
#
# Copyright (C) 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

'''
'''

from __future__ import print_function
import json
import requests
import sys

if len(sys.argv) != 3:
    sys.stderr.write('Usage: %s <Project> '
                     '<Upstream Change id>\n' % sys.argv[0])
    sys.exit(1)

r = requests.get('http://review.rdoproject.org/r'
                 "/changes/?q=status:open+project:%s&o=CURRENT_COMMIT&"
                 "o=CURRENT_REVISION" % sys.argv[1])

changes = json.loads('\n'.join(r.text.split('\n')[1:]))

for change in changes:
    for rev in change['revisions'].keys():
        lines = change['revisions'][rev]['commit']['message'].split('\n')
        for line in lines:
            parts = line.strip().split(': ')
            if (len(parts) == 2 and parts[0] == 'Upstream-Id' and
               parts[1] == sys.argv[2]):
                print(change['change_id'])
                sys.exit(0)
sys.exit(1)

# get_rdo_review.py ends here
